"""
Implementation of linear_composite() function as a module.
"""

import numpy as np
import pandas as pd
from typing import List, Dict, Union, Optional, Tuple, Type, Set
import warnings
from packaging import version
from macrosynergy.management.utils import reduce_df, is_valid_iso_date
from macrosynergy.management.simulate import make_test_df
from macrosynergy.management.types import QuantamentalDataFrame

listtypes: Tuple[Type, ...] = (list, np.ndarray, pd.Series, tuple)

PD_FUTURE_STACK = (
    dict(future_stack=True)
    if version.parse(pd.__version__) > version.parse("2.1.0")
    else dict(dropna=False)
)


def linear_composite(
    df: pd.DataFrame,
    xcats: Union[str, List[str]],
    cids: Optional[List[str]] = None,
    weights: Optional[Union[List[float], str]] = None,
    normalize_weights: bool = True,
    signs: Optional[List[float]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Dict[str, List[str]] = None,
    complete_xcats: bool = False,
    complete_cids: bool = False,
    new_xcat="NEW",
    new_cid="GLB",
):
    """
    Weighted linear combinations of cross sections or categories

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized JPMaQS DataFrame with the necessary columns: 'cid', 'xcat',
        'real_date' and 'value'.
    xcats : Union[str, List[str]
        One or more categories to be combined. If a single category is given the linear
        combination is calculated across cross-sections. This results in a single series
        to which a new cross-sectional identifier is assigned. If more than one category
        string is given the output will be a new category, i.e. a panel that is a linear
        combination of the categories specified.
    cids : List[str]
        cross-sections for which the linear combinations are calculated. Default is all
        cross-section available.
    weights : Union[List[float], str]
        This specifies how categories or cross sections are combined. There are three
        principal options. The first (default) is None, in which case equal weights are
        given to all categories or cross sections that are available. The second case is a
        set of fixed coefficients, in which case these very coefficients are applied to all
        available categories of cross sections. Per default the coefficients are normalized
        so that they add up to one for each period. This can be changed with the argument
        `normalize_weights`. The third case is the assignment of a weighting category. This
        only applies to combinations of cross sections. In this case the weighting category
        is multiplied for each period with the corresponding value of main category of the
        same cross section. Per default the weight category values are normalized so that
        they add up to one for each period. This can be changed with the argument
        `normalize_weights`.
    normalize_weights : bool
        If True (default) the weights are normalized to sum to 1. If False the weights
        are used as specified.
    signs : List[float]
        An array of consisting of +1s or -1s, of the same length as the number of
        categories in `xcats` to indicate whether the respective category should be added or
        subtracted from the linear combination. Not relevant when aggregating over cross-
        sections, i.e. when a single category is given in `xcats`. Default is None and all
        signs are set to +1.
    start : str
        earliest date in ISO format. Default is None and earliest date for which the
        respective category is available is used.
    end : str
        latest date in ISO format. Default is None and latest date for which the
        respective category is available is used.
    complete_xcats : bool
        If True (default) combinations are only calculated for observation dates on
        which all categories are available. If False a combination of the available
        categories is used. Not relevant when aggregating over cross-sections, i.e.
        when a single category is given in `xcats`.
    complete_cids : bool
        If True (default) combinations are only calculated for observation dates on
        which all cross-sections are available. If False a combination of the available
        cross-sections is used. Not relevant when aggregating over categories, i.e. when
        multiple categories are given in `xcats`.
    new_xcat : str
        Name of new composite category when aggregating over categories for a given
        cross-section. Default is "NEW".
    new_cid : str
        Name of new composite cross-section when aggregating over cross-sections for a 
        given category. Default is "GLB".

    Returns
    -------
    ~pandas.DataFrame
        standardized DataFrame with the composite values, with the columns:
        'cid', 'xcat', 'real_date' and 'value'.
    """

    (
        df,
        xcats,
        cids,
        weights,
        normalize_weights,
        signs,
        start,
        end,
        blacklist,
        complete_xcats,
        complete_cids,
        new_xcat,
        new_cid,
        _xcat_agg,
        mode,
    ) = _check_args(
        df=df,
        xcats=xcats,
        cids=cids,
        weights=weights,
        normalize_weights=normalize_weights,
        signs=signs,
        start=start,
        end=end,
        blacklist=blacklist,
        complete_xcats=complete_xcats,
        complete_cids=complete_cids,
        new_xcat=new_xcat,
        new_cid=new_cid,
    )

    # update local variables

    _xcats: List[str] = xcats + ([weights] if isinstance(weights, str) else [])

    remaining_xcats: List[str]
    remaining_cids: List[str]
    # NOTE: the "remaining_*" variables will not be in the same order as the input
    # cids/xcats.
    # Do not used these for index based lookups/operations.
    df, remaining_xcats, remaining_cids = reduce_df(
        df=df,
        xcats=_xcats,
        cids=cids,
        start=start,
        end=end,
        blacklist=blacklist,
        intersect=False,
        out_all=True,
    )

    df = QuantamentalDataFrame(df)
    result_as_categorical = df.InitializedAsCategorical

    if (
        len(remaining_cids) < len(cids)
        and not _xcat_agg
        and complete_cids
        or len(remaining_cids) == 0
    ):
        missing_cids_xcats_str = _missing_cids_xcats_str(df=df, cids=cids, xcats=xcats)
        raise ValueError(
            "Not all `cids` have complete `xcat` data required for the calculation.\n"
            f"{missing_cids_xcats_str}"
        )

    if _xcat_agg:
        df = _populate_missing_xcat_series(df)

        result_df: QuantamentalDataFrame = linear_composite_xcat_agg(
            df=df,
            xcats=xcats,
            weights=weights,
            signs=signs,
            normalize_weights=normalize_weights,
            complete_xcats=complete_xcats,
            new_xcat=new_xcat,
        )

    else:  # mode == "cid_agg" -- single xcat
        df, cids, _xcat, weights, signs = _check_df_for_missing_cid_data(
            df=df, cids=cids, weights=weights, signs=signs
        )

        result_df: QuantamentalDataFrame = linear_composite_cid_agg(
            df=df,
            xcat=_xcat,
            cids=cids,
            weights=weights,
            signs=signs,
            normalize_weights=normalize_weights,
            complete_cids=complete_cids,
            new_cid=new_cid,
        )
    
    return QuantamentalDataFrame(result_df, categorical=result_as_categorical)

def _missing_cids_xcats_str(
    df: QuantamentalDataFrame,
    cids: List[str],
    xcats: List[str],
) -> str:
    output_strs: List[str] = []

    found_cids = df["cid"].unique().tolist()
    found_xcats = df["xcat"].unique().tolist()

    if set(cids) != set(found_cids):
        missing_cids = list(set(cids) - set(found_cids))
    else:
        missing_cids = []

    if set(xcats) != set(found_xcats):
        missing_xcats = list(set(xcats) - set(found_xcats))
    else:
        missing_xcats = []

    xcat_dict: Dict[str, str] = {}
    for xc in sorted(xcats):
        miss_cids = list(
            set(cids) - set(df.loc[df["xcat"] == xc, "cid"].unique().tolist())
        )
        if miss_cids:
            xcat_dict[xc] = miss_cids

    if missing_cids:
        output_strs.append(f"Missing cids: {missing_cids}")
    if missing_xcats:
        output_strs.append(f"Missing xcats: {missing_xcats}")

    if xcat_dict:
        output_strs.append(
            "The following `cids` are missing for the respective `xcats`:"
        )
        longest_xc = max([len(xc) for xc in xcat_dict.keys()])
        for _xc, _cids in xcat_dict.items():
            msg = f"{_xc}: " + " " * (longest_xc - len(_xc)) + " " + str(sorted(_cids))
            output_strs.append(msg)

    return "\n".join(output_strs)


def _linear_composite_basic(
    data_df: pd.DataFrame,
    weights_df: pd.DataFrame,
    normalize_weights: bool = True,
    complete: bool = False,
    mode: str = "xcat_agg",
):
    """Main calculation function for linear_composite()"""

    # Create a boolean mask to help us work out the calcs
    nan_mask: pd.DataFrame = data_df.isna() | weights_df.isna()

    # Normalize weights (if requested)
    if normalize_weights:
        adj_weights_wide = weights_df[~nan_mask].div(
            weights_df[~nan_mask].abs().sum(axis=1), axis=0
        )
        adj_weights_wide[nan_mask] = np.nan

        assert np.allclose(
            adj_weights_wide[~adj_weights_wide.isna().all(axis=1)].abs().sum(axis=1), 1
        ), "Weights do not sum to 1. Normalization failed."

        weights_df = adj_weights_wide.copy()

    # Multiply the weights by the target data
    out_df = data_df * weights_df

    # Sum across the columns
    out_df = out_df.sum(axis="columns")

    # NOTE: Using `axis` with strings, to make it more readable
    # Remove periods with missing data (if requested) (rows with any NaNs)
    if complete:
        out_df[nan_mask.any(axis="columns")] = np.nan

    # put NaNs back in, as sum() removes them
    out_df[nan_mask.all(axis="columns")] = np.nan

    # Reset index, rename columns and return
    out_df = out_df.reset_index().rename(columns={0: "value"})

    # TODO: out_df from cid_agg and xcat_agg are not in the same format...

    return out_df


def linear_composite_cid_agg(
    df: QuantamentalDataFrame,
    xcat: str,
    cids: List[str],
    weights: Union[str, List[float]],
    signs: List[float],
    normalize_weights: bool = True,
    complete_cids: bool = True,
    new_cid="GLB",
):
    """Linear composite of various cids for a given category across all periods."""
    if isinstance(weights, str):
        weights_df: pd.DataFrame = df[(df["xcat"] == weights)]
        weights_df = weights_df.set_index(["real_date", "cid"])["value"].unstack(
            level=1
        )
        weights_df = weights_df[cids].mul(signs, axis=1)

    else:
        weights_series: pd.Series = pd.Series(
            np.array(weights) * np.array(signs),
            index=cids,
        )
        weights_df = pd.DataFrame(
            data=[weights_series.sort_index()],
            index=pd.to_datetime(df["real_date"].unique().tolist()),
            columns=df["cid"].unique(),
        )

        weights_df.index.names = ["real_date"]
        weights_df.columns.names = ["cid"]

    # create the data_df
    data_df: pd.DataFrame = (
        df[(df["xcat"] == xcat)]
        .set_index(["real_date", "cid"])["value"]
        .unstack(level=1)
    )
    # aligning the index of weights_df to the data one
    # so that we have the same set of dates and same set of CIDs -- thank you
    # @mikiinterfiore
    weights_df = (
        weights_df.stack(**PD_FUTURE_STACK)
        .reindex(data_df.stack(**PD_FUTURE_STACK).index)
        .unstack(1)
    )

    # assert that data_df and weights_df have the same shape, index and columns
    assert (
        (data_df.shape == weights_df.shape)
        and (data_df.index.equals(weights_df.index))
        and (data_df.columns.equals(weights_df.columns))
    ), (
        "Unexpected shape of `data_df` and `weights_df`. "
        "Unable to shape data for calculation."
    )

    # Calculate the linear combination
    out_df: pd.DataFrame = _linear_composite_basic(
        data_df=data_df,
        weights_df=weights_df,
        normalize_weights=normalize_weights,
        complete=complete_cids,
        mode="cid_agg",
    )

    if df.is_categorical():
        out_df = QuantamentalDataFrame.from_timeseries(
            out_df.set_index("real_date")["value"], ticker=f"{new_cid}_{xcat}"
        )
    else:
        out_df["cid"] = new_cid
        out_df["xcat"] = xcat

    return out_df


def linear_composite_xcat_agg(
    df: QuantamentalDataFrame,
    xcats: List[str],
    weights: List[float],
    signs: List[float],
    normalize_weights: bool = True,
    complete_xcats: bool = True,
    new_xcat="NEW",
):
    """Linear composite of various xcats across all cids and periods"""

    # Create a weights series with the xcats as index
    weights_series: pd.Series = pd.Series(
        np.array(weights) * np.array(signs), index=xcats
    )

    # Create wide dataframes for the data and weights
    data_df = df.set_index(["cid", "real_date", "xcat"])["value"].unstack(level=2)
    weights_df = pd.DataFrame(
        data=[weights_series.sort_index()],
        index=data_df.index,
        columns=data_df.columns,
    )

    # Calculate the linear combination
    out_df: pd.DataFrame = _linear_composite_basic(
        data_df=data_df,
        weights_df=weights_df,
        normalize_weights=normalize_weights,
        complete=complete_xcats,
        mode="xcat_agg",
    )
    if df.is_categorical():
        # add a new column called xcat with the new_xcat value
        out_df["xcat"] = pd.Categorical.from_codes(
            codes=[0] * len(out_df), categories=[new_xcat]
        )
        out_df = QuantamentalDataFrame(out_df)
    else:
        out_df["xcat"] = new_xcat

    return out_df


def _populate_missing_xcat_series(
    df: QuantamentalDataFrame,
) -> QuantamentalDataFrame:
    """
    Populate missing xcat series with NaNs
    """
    found_cids: List[str] = df["cid"].unique().tolist()
    found_xcats: List[str] = df["xcat"].unique().tolist()
    found_xcats_set: Set[str] = set(found_xcats)
    dt_range: pd.DatetimeIndex = pd.to_datetime(df["real_date"].unique())
    wrn_msg: str = (
        "{cidx} does not have complete xcat data for {missing_xcats}."
        " These will be filled with NaNs for the calculation."
    )

    for cidx in found_cids:
        missing_xcats = list(
            found_xcats_set - set(df.loc[df["cid"] == cidx, "xcat"].unique())
        )
        if missing_xcats:
            warnings.warn(wrn_msg.format(cidx=cidx, missing_xcats=missing_xcats))
            for xc in missing_xcats:
                if df.is_categorical():
                    df.add_nan_series(
                        ticker=f"{cidx}_{xc}",
                        start=dt_range.min(),
                        end=dt_range.max(),
                    )
                else:
                    dct = {
                        "cid": cidx,
                        "xcat": xc,
                        "real_date": dt_range,
                        "value": np.nan,
                    }
                    df = pd.concat([df, pd.DataFrame(data=dct)])

    return df


def _check_df_for_missing_cid_data(
    df: QuantamentalDataFrame,
    cids: List[str],
    weights: Union[str, List[float]],
    signs: List[float],
) -> Tuple[
    QuantamentalDataFrame, List[str], str, Union[str, List[float], None], List[float]
]:
    """
    Check the DataFrame for missing `cid` data and drop them if necessary and return the
    DataFrame with the missing `cid` data dropped.
    """

    found_cids: List[str] = df["cid"].unique().tolist()
    found_cids = [cid for cid in cids if cid in found_cids]
    found_xcats: List[str] = df["xcat"].unique().tolist()
    found_xcats_set: Set[str] = set(found_xcats)
    wrn_msg: str = (
        "`cid` {cidx} does not have complete `xcat` data for {missing_xcats}."
        " These will be dropped from the calculation."
    )
    if isinstance(weights, str):
        if weights not in found_xcats:
            raise ValueError(
                f"Weight category {weights} not found in `df`. "
                f"Available categories are {found_xcats}."
            )

        if len(found_xcats_set - {weights}) == 0:
            raise ValueError(
                "None of the `xcats` are present in `df` other than the `weights`. "
                f"Available categories are {found_xcats}."
            )

    if set(cids) - set(found_cids) != set():
        for cid in set(cids) - set(found_cids):
            # Cids has already been removed since it uses
            warnings.warn(f"cid {cid} not found in `df`. It will be ignored.")
            signs.pop(cids.index(cid))
            if isinstance(weights, list):
                weights.pop(cids.index(cid))

    ctr = 0
    for cidx in found_cids.copy():  # copy to allow modification of `cids`
        missing_xcats = list(
            found_xcats_set - set(df.loc[df["cid"] == cidx, "xcat"].unique())
        )
        if missing_xcats:
            found_cids.pop(ctr)
            signs.pop(ctr)
            if isinstance(weights, list):
                weights.pop(ctr)
            # drop from df
            df = df.loc[df["cid"] != cidx, :]
            warnings.warn(wrn_msg.format(cidx=cidx, missing_xcats=missing_xcats))
        else:
            ctr += 1

    if len(found_cids) == 0:
        raise ValueError(
            "No `cids` have complete `xcat` data required for the calculation."
        )

    _xcat: str = list(set(found_xcats) - {weights if isinstance(weights, str) else ""})[
        0
    ]

    rcids = [c for c in cids if c in found_cids]  # to preserve order
    return QuantamentalDataFrame(df), rcids, _xcat, weights, signs

def _check_args(
    df: QuantamentalDataFrame,
    xcats: Union[str, List[str]],
    cids: Optional[List[str]] = None,
    weights: Optional[Union[List[float], str]] = None,
    normalize_weights: bool = True,
    signs: Optional[List[float]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Dict[str, List[str]] = None,
    complete_xcats: bool = False,
    complete_cids: bool = False,
    new_xcat="NEW",
    new_cid="GLB",
):
    """
    Check the arguments of linear_composite()
    """

    # df check
    if (
        (not isinstance(df, QuantamentalDataFrame))
        or ("value" not in df.columns)
        or (df["value"].isna().all())
    ):
        raise TypeError("`df` must be a standardized Quantamental DataFrame.")

    if start is None:
        start: str = df["real_date"].min().strftime("%Y-%m-%d")
    if end is None:
        end: str = df["real_date"].max().strftime("%Y-%m-%d")

    # dates check
    for varx, namex in zip([start, end], ["start", "end"]):
        if varx is not None:
            if not (isinstance(varx, str) and is_valid_iso_date(varx)):
                raise ValueError(f"`{namex}` must be a valid ISO date string.")

    # if type(df) is QuantamentalDataFrame and df.is_categorical():
    #     xcats_in_df = set(df["xcat"].cat.categories)
    #     cids_in_df = set(df["cid"].cat.categories)
    # else:
    xcats_in_df = set(df["xcat"].values)
    cids_in_df = set(df["cid"].values)

    # check xcats
    if xcats is None:
        xcats: List[str] = list(xcats_in_df)
    elif isinstance(xcats, str):
        xcats: List[str] = [xcats]
    elif isinstance(xcats, listtypes):
        xcats: List[str] = list(xcats)
    else:
        raise TypeError("`xcats` must be a string or list of strings.")

    if not all(x in xcats_in_df for x in xcats):
        if complete_xcats:
            raise ValueError("Not all `xcats` are available in `df`.")
        else:
            missing_xcats = list(set(xcats) - xcats_in_df)
            warnings.warn(
                f"Not all `xcats` are available in `df`: {missing_xcats} "
                "The calculation will be performed with the available xcats."
            )
            if signs is not None:
                signs = [signs[i] for i, xc in enumerate(xcats) if xc not in missing_xcats]
            if isinstance(weights, list):
                weights = [weights[i] for i, xc in enumerate(xcats) if xc not in missing_xcats]
            xcats = [xc for xc in xcats if xc not in missing_xcats]

    # check cids
    
    if cids is None:
        cids: List[str] = list(cids_in_df)
    elif isinstance(cids, str):
        cids: List[str] = [cids]
    elif isinstance(cids, listtypes):
        cids: List[str] = list(cids)
    else:
        raise TypeError("`cids` must be a string or list of strings.")

    # check cids in df
    if not all(c in cids_in_df for c in cids):
        if complete_cids:
            raise ValueError("Not all `cids` are available in `df`.")
        else:
            missing_cids = list(set(cids) - cids_in_df)
            warnings.warn(
                f"Not all `cids` are available in `df`: {missing_cids} "
                "The calculation will be performed with the available cids."
            )
            if signs is not None:
                signs = [signs[i] for i, cid in enumerate(cids) if cid not in missing_cids]
            if isinstance(weights, list):
                weights = [weights[i] for i, cid in enumerate(cids) if cid not in missing_cids]
            cids = [cid for cid in cids if cid not in missing_cids]

    _xcat_agg: bool = len(xcats) > 1 or new_xcat != "NEW"
    mode: str = "xcat_agg" if _xcat_agg else "cid_agg"

    if _xcat_agg and isinstance(weights, str):
        raise ValueError(
            "When aggregating over xcats, `weights` "
            "must be a list of floats or integers."
        )

    # check weights
    expc_weights_len: int = len(xcats) if _xcat_agg else len(cids)

    if weights is None:
        weights: List[float] = list(np.ones(expc_weights_len) / expc_weights_len)
    elif isinstance(weights, listtypes):
        weights: List[float] = list(weights)
        if not all([isinstance(x, (float, int)) for x in weights]):
            raise TypeError("`weights` must be a list of floats or integers.")
        if len(weights) != expc_weights_len:
            raise ValueError(
                "`weights` must be a list of floats of the same length as `xcats`."
            )
        if any([x == 0.0 for x in weights]):
            raise ValueError("`weights` must not contain any 0s.")

    elif isinstance(weights, str):
        if weights not in xcats_in_df:
            raise ValueError(
                "When using a category-string as `weights`"
                " it must be present in `df`."
            )
    else:
        raise TypeError("`weights` must be a list of floats, a string or None.")

    # check signs
    if signs is None:
        signs: List[float] = [1.0] * (len(xcats) if _xcat_agg else len(cids))
    elif isinstance(signs, listtypes):
        signs: List[float] = list(signs)
        if len(signs) != expc_weights_len:
            raise ValueError(
                "`signs` must be a list of floats of the same length as `xcats`."
            )
        if not all([x in [-1.0, 1.0] for x in signs]):
            if any([x == 0.0 for x in signs]):
                raise ValueError("`signs` must not contain any 0s.")
            warnings.warn(
                "`signs` must be a list of +1s or -1s. "
                "`signs` will be coerced to +1s/-1s. "
                "(i.e. signs = abs(signs) / signs)"
            )

            signs: List[float] = [abs(x) / x for x in signs]

    else:
        raise TypeError("`signs` must be a list of floats/ints or None.")

    if not isinstance(normalize_weights, bool):
        raise TypeError("`normalize_weights` must be a boolean.")

    if not isinstance(complete_xcats, bool):
        raise TypeError("`complete_xcats` must be a boolean.")

    if not isinstance(complete_cids, bool):
        raise TypeError("`complete_cids` must be a boolean.")

    if not isinstance(new_xcat, str):
        raise TypeError("`new_xcat` must be a string.")

    if not isinstance(new_cid, str):
        raise TypeError("`new_cid` must be a string.")

    if blacklist is not None:
        if not isinstance(blacklist, dict):
            raise TypeError("`blacklist` must be a dictionary.")

    return (
        df,
        xcats,
        cids,
        weights,
        normalize_weights,
        signs,
        start,
        end,
        blacklist,
        complete_xcats,
        complete_cids,
        new_xcat,
        new_cid,
        _xcat_agg,
        mode,
    )


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP"]
    xcats = ["XR", "CRY", "INFL"]

    df: pd.DataFrame = pd.concat(
        [
            make_test_df(
                cids=cids,
                xcats=xcats[:-1],
                start="2000-01-01",
                end="2000-02-01",
                style="linear",
            ),
            make_test_df(
                cids=cids,
                xcats=["INFL"],
                start="2000-01-01",
                end="2000-02-01",
                style="decreasing-linear",
            ),
        ]
    )

    # all infls are now decreasing-linear, while everything else is increasing-linear

    df.loc[
        (df["cid"] == "GBP")
        & (df["xcat"] == "INFL")
        & (df["real_date"] == "2000-01-17"),
        "value",
    ] = np.nan

    df.loc[
        (df["cid"] == "AUD")
        & (df["xcat"] == "CRY")
        & (df["real_date"] == "2000-01-17"),
        "value",
    ] = np.nan

    # there are now missing values for AUD-CRY and GBP-INFL on 2000-01-17

    lc_cid = linear_composite(
        df=df, xcats="XR", weights="INFL", normalize_weights=False
    )
    df = QuantamentalDataFrame(df)
    lc_xcat = linear_composite(
        df=df,
        cids=["GBP", "AUD", "CAD"],
        xcats=["XR"],
        weights=[1, 2, 1],
        signs=[1, -1, 1],
        complete_xcats=True,
    )


import numpy as np
import pandas as pd

import sklearn.ensemble as skl
from sklearn.utils import Bunch

class VotingRegressor(skl.VotingRegressor):
    """
    Regression model that averages the predictions of many regression models.

    Parameters
    ----------
    estimators : list of (str, estimator) tuples
        List of (name, estimator) tuples that are used to fit the model.
    weights : array-like of shape (n_estimators,), default=None
        Sequence of weights to assign to models. If None, models are weighted
        equally.
    verbose : bool, default=False
        If True, the time elapsed while fitting will be printed as model
        trains.

    Notes
    -----
    This class calculates feature importances as the average of the feature
    importances of the base estimators.
    """
    def __init__(self, estimators, weights=None, verbose=False):
        super().__init__(estimators=estimators, weights=weights, n_jobs=1, verbose=verbose)

        # Attributes
        self.feature_importances_ = None
        self.estimators_ = None
        self.named_estimators_ = None
        
    def fit(self, X, y, **fit_params):
        """
        Fit the estimators.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            Pandas dataframe or numpy array of input features.
        y : pd.Series or pd.DataFrame or np.ndarray
            Pandas series, dataframe or numpy array of targets associated with each sample
            in X.
        """
        # Checks 
        self._check_fit_params(X, y)

        # Fit regressors
        self.estimators_ = []
        self.named_estimators_ = Bunch()

        importances = []
        for _, estimator in self.estimators:
            # Fit the estimator
            self.estimators_.append(estimator.fit(X, y, **fit_params))
            # Store feature importances if available
            if hasattr(estimator, "coef_") or hasattr(estimator, "feature_importances_"):
                # Normalize feature importances to sum to 1
                imp = (
                    np.abs(estimator.coef_) / np.sum(np.abs(estimator.coef_))
                    if hasattr(estimator, "coef_")
                    else estimator.feature_importances_
                    / np.sum(estimator.feature_importances_)
                )
                importances.append(imp)

        if len(importances) > 0:
            if self.weights:
                self.feature_importances_ = np.average(
                    importances, weights=self.weights, axis=0
                )
            else:
                self.feature_importances_ = np.mean(importances, axis=0)

        # Renormalize feature importances to sum to 1
        if self.feature_importances_ is not None:
            self.feature_importances_ /= np.sum(self.feature_importances_)

        # Store named estimators
        self.named_estimators_ = Bunch(
            **{name: estimator for name, estimator in self.estimators}
        )

        return self

    
    def _check_fit_params(self, X, y):
        """
        Checks for fit method parameters
        """
        # X
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Input feature matrix for the voting regressor must be either a pandas "
                "dataframe or numpy array."
            )
        if isinstance(X, np.ndarray):
            if X.ndim != 2:
                raise ValueError(
                    "When the input feature matrix for the voting regressor is a "
                    "numpy array, it must have two dimensions."
                )
        # y
        if not isinstance(y, (pd.Series, pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Target vector for the voting regressor must be either a pandas series, "
                "dataframe or numpy array."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The dependent variable dataframe must have only one column. If used "
                    "as part of an sklearn pipeline, ensure that previous steps return "
                    "a pandas series or dataframe."
                )
        if isinstance(y, np.ndarray):
            if y.ndim != 1:
                raise ValueError(
                    "When the target vector for the voting regressor is a numpy "
                    "array, it must have one dimension."
                )
        if X.shape[0] != y.shape[0]:
            raise ValueError(
                "The number of samples in the input feature matrix must match the number "
                "of samples in the target vector."
            )
    
class VotingClassifier(skl.VotingClassifier):
    """
    Classification model that votes on the predictions of many classifiers.

    Parameters
    ----------
    estimators : list of (str, estimator) tuples
        List of (name, estimator) tuples that are used to fit the model.
    voting : {'hard', 'soft'}, default='hard'
        If 'hard', uses predicted class labels for majority rule voting.
        If 'soft', predicts the class label based on the argmax of the sums of
        the predicted probabilities, which is recommended for an ensemble of
        well-calibrated classifiers.
    weights : array-like of shape (n_estimators,), default=None
        Sequence of weights to assign to models. If None, models are weighted
        equally.
    n_jobs : int, default=None
        The number of jobs to run in parallel for `fit`. `None` means 1 unless
        in a `joblib.parallel_backend` context. `-1` means using all
        processors.
    flatten_transform : bool, default=True
        Affects shape of transform output only when voting='soft'. If True,
        the transform method returns a matrix with shape (n_samples, n_classes*n_classifiers).
        If False, the shape is (n_classifiers, n_samples, n_classes).
    verbose : bool, default=False
        If True, the time elapsed while fitting will be printed as model
        trains.

    Notes
    -----
    This class calculates feature importances as the average of the feature
    importances of the base estimators.
    """
    def __init__(self, estimators, voting="hard", weights=None, n_jobs=None, flatten_transform=True, verbose=False):
        super().__init__(estimators=estimators, voting=voting, weights=weights, n_jobs=n_jobs, flatten_transform=flatten_transform, verbose=verbose)
        self.feature_importances_ = None

    def fit(self, X, y, sample_weight=None, **fit_params):
        """
        Fit the estimators.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            Pandas dataframe or numpy array of input features.
        y : pd.Series or pd.DataFrame or np.ndarray
            Pandas series, dataframe or numpy array of targets associated with each sample
            in X.
        """
        # Checks
        self._check_fit_params(X, y)

        # Fit classifiers
        super().fit(X, y, sample_weight, **fit_params)

        # Calculate feature importances
        importances = []
        for estimator in self.estimators_:
            if hasattr(estimator, "coef_") or hasattr(
                estimator, "feature_importances_"
            ):
                # Normalize feature importances to sum to 1
                imp = (
                    np.squeeze(np.abs(estimator.coef_) / np.sum(np.abs(estimator.coef_)))
                    if hasattr(estimator, "coef_")
                    else estimator.feature_importances_
                    / np.sum(estimator.feature_importances_)
                )
                importances.append(imp)

        if len(importances) > 0:
            self.feature_importances_ = np.mean(importances, axis=0)
        
        # Renormalize feature importances to sum to 1
        if self.feature_importances_ is not None:
            self.feature_importances_ /= np.sum(self.feature_importances_)

        return self
    
    def _check_fit_params(self, X, y):
        """
        Checks for fit method parameters
        """
        # X
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Input feature matrix for the voting regressor must be either a pandas "
                "dataframe or numpy array."
            )
        if isinstance(X, np.ndarray):
            if X.ndim != 2:
                raise ValueError(
                    "When the input feature matrix for the voting regressor is a "
                    "numpy array, it must have two dimensions."
                )
        # y
        if not isinstance(y, (pd.Series, pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Target vector for the voting regressor must be either a pandas series, "
                "dataframe or numpy array."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The dependent variable dataframe must have only one column. If used "
                    "as part of an sklearn pipeline, ensure that previous steps return "
                    "a pandas series or dataframe."
                )
        if isinstance(y, np.ndarray):
            if y.ndim != 1:
                raise ValueError(
                    "When the target vector for the voting regressor is a numpy "
                    "array, it must have one dimension."
                )
        if X.shape[0] != y.shape[0]:
            raise ValueError(
                "The number of samples in the input feature matrix must match the number "
                "of samples in the target vector."
            )


if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf
    import pandas as pd

    from sklearn.linear_model import LinearRegression, LogisticRegression
    from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()

    # Regressor
    X_train = train.drop(columns=["XR"])
    y_train = train["XR"]

    vr = VotingRegressor(
        estimators = [
            ("lr", LinearRegression()),
            ("rf", RandomForestRegressor(random_state = 42))
        ],
        weights = [0.3, 0.7],
    ).fit(X_train, y_train)
    print(vr.weights)
    print(f"Voting regressor feature importances: {vr.feature_importances_}")

    # Classifier
    X_train = train.drop(columns=["XR"])
    y_train = np.sign(train["XR"])

    vr = VotingClassifier(
        estimators = [
            ("lr", LogisticRegression()),
            ("rf", RandomForestClassifier())
        ]
    ).fit(X_train, y_train)

    print(f"Voting classifier feature importances: {vr.feature_importances_}")


from abc import ABC, abstractmethod
import pandas as pd

from macrosynergy.management.types.qdf.classes import QuantamentalDataFrame


class BaseTransformer(ABC):
    @abstractmethod
    def transform(self, data, **kwargs):
        pass

    @staticmethod
    def get_mapping_value(d, keys):
        if not isinstance(keys, list):
            return d[keys]

        for key in keys:
            d = d[key]
        return d


class DataFrameTransformer(BaseTransformer):
    """
    This transformer is used to transform a pandas DataFrame into a QuantamentalDataFrame.
    """
    def transform(self, data, mapping):
        if data.empty:
            raise ValueError("DataFrame is empty.")
        if not isinstance(mapping, dict):
            raise TypeError("Mapping must be a dictionary.")

        # Case 1 - Mapping is a dictionary of columns to rename.
        if set(mapping.keys()).issubset(data.columns):
            df = data.rename(columns=mapping).drop(
                columns=set(data.columns) - set(mapping.keys())
            )
            return QuantamentalDataFrame(df)

        # Case 2 - Dataframe has some multiindexing in the columns
        elif isinstance(data.columns, pd.MultiIndex):
            df = data.stack(level=[0, 1]).reset_index()
            if "value" not in mapping.values():
                mapping[0] = "value"
            df.rename(columns=mapping, inplace=True)
            return QuantamentalDataFrame(df)

        # Case 3 - Dataframe has some multiindexing in the indexes
        elif isinstance(data.index, pd.MultiIndex):
            df = data.stack(level=0).reset_index()
            if "value" not in mapping.values():
                mapping[0] = "value"
            df.rename(columns=mapping, inplace=True)
            return QuantamentalDataFrame(df)
        
        # Case 4 - Need to split into xcat and cid
        else:
            raise ValueError(
                "Mapping keys must be a subset of the DataFrame columns and names" 
                " of indexes/multiindexes."
            )

class JSONTransformer(BaseTransformer):
    """
    This transformer is used to transform a list of JSON objects into a QuantamentalDataFrame.

    Any JSON object is assumed to be a dictionary in python.
    """
    def transform(self, data, mapping):

        rows = []

        for json in data:
            cid = BaseTransformer.get_mapping_value(json, mapping["cid"])
            xcat = BaseTransformer.get_mapping_value(json, mapping["xcat"])
            time_series = BaseTransformer.get_mapping_value(json, mapping["time-series"])
            for date, value in time_series:
                rows.append({"real_date": date, "cid": cid, "xcat": xcat, "value": value})

        df = pd.DataFrame(rows)
        df["real_date"] = pd.to_datetime(df["real_date"])

        return QuantamentalDataFrame(df)

def transform_to_qdf(data, **kwargs):
    if isinstance(data, pd.DataFrame):
        return DataFrameTransformer().transform(
            data, mapping=kwargs.get("mapping", {})
        )
    elif isinstance(data, list) and all(isinstance(x, dict) for x in data):
        return JSONTransformer().transform(
            data, mapping=kwargs.get("mapping", {})
        )
    else:
        raise TypeError(
            "data format has not been implemented, please create a new transformer "
            "class that inherits from BaseTransformer and use the transform method "
            "directly."
        )


if __name__ == "__main__":
    df = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": ["s", "g", "q"],
            "c": ["baap", "boop", "beep"],
            "date": ["2021-01-01", "2021-01-02", "2021-01-03"],
        }
    )

    new_df = transform_to_qdf(
        df, mapping={"a": "value", "b": "cid", "c": "xcat", "date": "real_date"}
    )

    print(new_df)

    import os
    from macrosynergy.download import DataQueryInterface

    client_id: str = os.getenv("DQ_CLIENT_ID")
    client_secret: str = os.getenv("DQ_CLIENT_SECRET")

    expressions = ["DB(CFX,GBP,)", "DB(CFX,USD,)"]

    with DataQueryInterface(
        client_id=client_id,
        client_secret=client_secret,
    ) as dq:
        assert dq.check_connection(verbose=True)

        data = dq.download_data(
            expressions=expressions,
            start_date="2024-01-25",
            end_date="2024-02-05",
            show_progress=True,
        )

    new_df = transform_to_qdf(data, mapping = {
        "cid": ["attributes", 0, "expression"],
        "xcat": ["attributes", 0, "expression"],
        "time-series": ["attributes", 0, "time-series"]
    })

    print(new_df)



"""
Module for checking the availability of data availabity from a Quantamental DataFrame.
Includes functions for checking start years and end dates of a DataFrame, as well as
visualizing the results.
"""

import numpy as np
import pandas as pd
from typing import List, Tuple

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df
from macrosynergy.management.types import QuantamentalDataFrame
import macrosynergy.visuals as msv


def check_availability(
    df: pd.DataFrame,
    xcats: List[str] = None,
    cids: List[str] = None,
    start: str = None,
    start_size: Tuple[float] = None,
    end_size: Tuple[float] = None,
    start_years: bool = True,
    missing_recent: bool = True,
    use_last_businessday: bool = True,
    title: str = None,
    xcat_labels: dict = None,
    sort_labels: bool = False,
):
    """
    Wrapper for visualizing start and end dates of a filtered DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        standardized DataFrame with the following necessary columns: 'cid', 'xcat',
        'real_date'.
    xcats : List[str]
        extended categories to be checked on. Default is all in the DataFrame.
    cids : List[str]
        cross sections to be checked on. Default is all in the DataFrame.
    start : str
        string representing earliest considered date. Default is None, which reverts to
        earliest date in the dataframe.
    start_size : Tuple[float]
        tuple of floats with width / length of the start years heatmap. Default is None
        (format adjusted to data).
    end_size : Tuple[float]
        tuple of floats with width/length of the end dates heatmap. Default is None
        (format adjusted to data).
    start_years : bool
        boolean indicating whether or not to display a chart of starting years for each
        cross-section and indicator. Default is True (display start years).
    missing_recent : bool
        boolean indicating whether or not to display a chart of missing date numbers for
        each cross-section and indicator. Default is True (display missing days).
    use_last_businessday : bool
        boolean indicating whether or not to use the last business day before today as
        the end date. Default is True.
    title : str
        A string to be used as the title of the heatmap. If None, a default header will be
        used.
    xcat_labels : dict
        dictionary with xcat labels. Default is None (no labels).
    sort_labels : bool
        boolean indicating whether to sort the `xcats` in the heatmap alphabetically.
        The sorting is done based on the `xcats` list, with the labels from `xcat_labels`
        simply used for display (not regarded for sorting at all). Default is False (no
        sorting, ordered as provided in `xcats`).
    """

    for bvar, varname in zip(
        [start_years, missing_recent, sort_labels],
        ["start_years", "missing_recent", "sort_labels"],
    ):
        if not isinstance(bvar, bool):
            raise TypeError(f"`{varname}` must be a `bool` and not {type(bvar)}.")

    df = QuantamentalDataFrame(df)

    dfx = reduce_df(df, xcats=xcats, cids=cids, start=start)

    if xcats is None:
        xcats = sorted(dfx["xcat"].unique())

    if sort_labels:
        xcats = sorted(xcats)

    if xcat_labels is not None:
        dfx = dfx.rename_xcats(xcat_labels)

    if dfx.empty:
        raise ValueError(
            "No data available for the selected cross-sections and categories."
        )
    if start_years:
        dfs = check_startyears(dfx)
        row_order = get_heatmap_row_order(xcats=xcats, xcat_labels=xcat_labels)
        visual_paneldates(
            dfs,
            size=start_size,
            use_last_businessday=use_last_businessday,
            title=title,
            row_order=row_order,
        )
    if missing_recent:
        dfe = check_enddates(dfx)
        row_order = get_heatmap_row_order(xcats=xcats, xcat_labels=xcat_labels)
        visual_paneldates(
            dfe,
            size=end_size,
            use_last_businessday=use_last_businessday,
            title=title,
            row_order=row_order,
        )


def missing_in_df(
    df: QuantamentalDataFrame,
    xcats: List[str] = None,
    cids: List[str] = None,
):
    """
    Print missing cross-sections and categories

    Parameters
    ----------
    df : QuantamentalDataFrame
        standardized DataFrame with the following necessary columns: 'cid', 'xcat',
        'real_date'.
    xcats : List[str]
        extended categories to be checked on. Default is all in the DataFrame.
    cids : List[str]
        cross sections to be checked on. Default is all in the DataFrame.
    """

    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("`df` must be a QuantamentalDataFrame/pd.DataFrame")

    if df.empty:
        raise ValueError("`df` is empty.")

    for lst, name in zip([xcats, cids], ["xcats", "cids"]):
        if (lst is not None) and not (
            isinstance(lst, list) and all(isinstance(x, str) for x in lst)
        ):
            raise TypeError(f"`{name}` should be a `List[str]` and not {type(lst)}.")

    missing_across_df = list(set(xcats) - set(df["xcat"]))
    if len(missing_across_df) > 0:
        print("Missing XCATs across DataFrame: ", missing_across_df)
    else:
        print("No missing XCATs across DataFrame.")

    cids = df["cid"].unique() if cids is None else cids
    xcats_used = sorted(list(set(xcats).intersection(set(df["xcat"]))))
    if len(xcats_used) == 0:
        print("No XCATs found in the DataFrame.")
        return

    max_xcat_len = max(map(len, xcats_used))
    for xcat in xcats_used:
        cids_xcat = df.loc[df["xcat"] == xcat, "cid"].unique()
        missing_cids = sorted(set(cids) - set(cids_xcat))
        msg = f"Missing cids for {xcat}: " + " " * (max_xcat_len - len(xcat))
        print(msg, missing_cids)


def check_startyears(df: pd.DataFrame) -> pd.DataFrame:
    """
    DataFrame with starting years across all extended categories and cross-sections

    Parameters
    ----------
    df : pd.DataFrame
        standardized DataFrame with the following necessary columns: 'cid', 'xcat',
        'real_date'.

    Returns
    -------
    pd.DataFrame
        DataFrame consisting of starting years for all series.
    """

    df: pd.DataFrame = df.copy()
    df = df.dropna(how="any")
    df_starts = (
        df[["cid", "xcat", "real_date"]].groupby(["cid", "xcat"], observed=True).min()
    )
    df_starts["real_date"] = pd.DatetimeIndex(df_starts.loc[:, "real_date"]).year

    return df_starts.unstack().loc[:, "real_date"].astype(int, errors="ignore")


def check_enddates(df: pd.DataFrame) -> pd.DataFrame:
    """
    DataFrame with end dates across all extended categories and cross sections.

    Parameters
    ----------
    df : pd.DataFrame
        standardized DataFrame with the following necessary columns: 'cid', 'xcat',
        'real_date'.

    Returns
    -------
    pd.DataFrame
        DataFrame consisting of end dates for all series.
    """

    df: pd.DataFrame = df.copy()
    df = df.dropna(how="any")
    df_ends = (
        df[["cid", "xcat", "real_date"]].groupby(["cid", "xcat"], observed=True).max()
    )
    df_ends["real_date"] = df_ends["real_date"].dt.strftime("%Y-%m-%d")

    return df_ends.unstack().loc[:, "real_date"]


def business_day_dif(df: pd.DataFrame, maxdate: pd.Timestamp) -> pd.DataFrame:
    """
    Number of business days between two respective business dates.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame cross-sections rows and category columns. Each cell in the DataFrame
        will correspond to the start date of the respective series.
    maxdate : pd.Timestamp
        maximum release date found in the received DataFrame. In principle, all series
        should have values up until the respective business date. The difference will
        represent possible missing values.

    Returns
    -------
    pd.DataFrame
        DataFrame consisting of business day differences for all series.
    """

    year_df = (maxdate.year - df.apply(lambda x: x.dt.isocalendar().year)) * 52
    week_df = maxdate.week - df.apply(lambda x: x.dt.isocalendar().week)
    # Account for difference over a year.
    week_df += year_df
    # Account for weekends.
    week_df *= 2
    df = (maxdate - df).apply(lambda x: x.dt.days)
    # set to zero if the difference is negative.
    df = df - week_df
    return df.where(df >= 0, 0)


def get_heatmap_row_order(xcats: List[str], xcat_labels: dict = None) -> List[str]:
    if not xcat_labels:
        return xcats
    missing = set(xcats) - set(xcat_labels.keys())
    if missing:
        raise ValueError(
            f"Missing labels for xcats: {sorted(missing)}. "
            "Ensure all specified `xcats` are present in the `xcat_labels` dictionary."
        )

    return [xcat_labels[xcat] for xcat in xcats] if xcat_labels else xcats


def visual_paneldates(
    df: pd.DataFrame,
    size: Tuple[float] = None,
    use_last_businessday: bool = True,
    title: str = None,
    row_order: List[str] = None,
):
    """
    Visualize panel dates with color codes.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame cross sections rows and category columns.
    size : Tuple[float]
        tuple of floats with width/length of displayed heatmap.
    use_last_businessday : bool
        boolean indicating whether or not to use the last business day before today as
        the end date. Default is True.
    title : str
        A string to be used as the title of the heatmap. If None, a default header will be
        used.
    row_order : List[str]
        A list of strings specifying the order of rows in the heatmap. These rows
        correspond to the columns of the input DataFrame. If None, the default order
        used by Seaborn will be applied.
    """

    msv.view_panel_dates(
        df=df,
        size=size,
        use_last_businessday=use_last_businessday,
        header=title,
        row_order=row_order,
    )


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP"]
    xcats = ["XR", "CRY"]

    cols_1 = ["earliest", "latest", "mean_add", "sd_mult"]
    df_cids = pd.DataFrame(index=cids, columns=cols_1)
    df_cids.loc["AUD",] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD",] = ["2010-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP",] = ["2012-01-01", "2020-11-30", -0.2, 0.5]

    cols_2 = cols_1 + ["ar_coef", "back_coef"]
    df_xcats = pd.DataFrame(index=xcats, columns=cols_2)
    df_xcats.loc["XR",] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["CRY",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    filt_na = (dfd["cid"] == "CAD") & (dfd["real_date"] < "2011-01-01")
    dfd.loc[filt_na, "value"] = np.nan

    xxcats = xcats + ["TREND"]
    xxcids = cids + ["USD"]

    check_availability(
        df=dfd,
        xcats=xcats,
        cids=cids,
        start_size=(10, 5),
        end_size=(10, 8),
        xcat_labels={"XR": "Exchange Rate", "CRY": "Commodity"},
    )



"""
Interface for downloading data from the JPMorgan DataQuery API. This module is not
intended to be used directly, but rather through macrosynergy.download.jpmaqs.py.
However, for a use cases independent of JPMaQS, this module can be used directly to
download data from the JPMorgan DataQuery API.
"""

import concurrent.futures
import time
import os
import logging
import itertools
import base64
import uuid
import io
import warnings
import requests
from datetime import datetime, timedelta, timezone
from typing import List, Optional, Dict, Union, Tuple, Any
from tqdm import tqdm

from macrosynergy import __version__ as ms_version_info
from macrosynergy.download.exceptions import (
    AuthenticationError,
    DownloadError,
    InvalidResponseError,
    HeartbeatError,
    NoContentError,
    KNOWN_EXCEPTIONS,
)
from macrosynergy.management.utils import (
    is_valid_iso_date,
    form_full_url,
)

CERT_BASE_URL: str = "https://platform.jpmorgan.com/research/dataquery/api/v2"
OAUTH_BASE_URL: str = (
    "https://api-developer.jpmorgan.com/research/dataquery-authe/api/v2"
)
OAUTH_TOKEN_URL: str = "https://authe.jpmchase.com/as/token.oauth2"
OAUTH_DQ_RESOURCE_ID: str = "JPMC:URI:RS-06785-DataQueryExternalApi-PROD"
JPMAQS_GROUP_ID: str = "JPMAQS"
API_DELAY_PARAM: float = 0.2  # 300ms delay between requests
TOKEN_EXPIRY_BUFFER: float = 0.9  # 90% of token expiry time.
API_RETRY_COUNT: int = 5  # retry count for transient errors
HL_RETRY_COUNT: int = 5  # retry count for "high-level" requests
MAX_CONTINUOUS_FAILURES: int = 5  # max number of continuous errors before stopping
HEARTBEAT_ENDPOINT: str = "/services/heartbeat"
TIMESERIES_ENDPOINT: str = "/expressions/time-series"
CATALOGUE_ENDPOINT: str = "/group/instruments"
HEARTBEAT_TRACKING_ID: str = "heartbeat"
OAUTH_TRACKING_ID: str = "oauth"
TIMESERIES_TRACKING_ID: str = "timeseries"
CATALOGUE_TRACKING_ID: str = "catalogue"

logger: logging.Logger = logging.getLogger(__name__)
debug_stream_handler = logging.StreamHandler(io.StringIO())
debug_stream_handler.setLevel(logging.NOTSET)
debug_stream_handler.setFormatter(
    logging.Formatter(
        "%(asctime)s - %(levelname)s - %(module)s - %(funcName)s :: %(message)s"
    )
)
logger.addHandler(debug_stream_handler)


def validate_response(
    response: requests.Response,
    user_id: str,
) -> dict:
    """
    Validates a response from the API. Raises an exception if the response is invalid
    (e.g. if the response is not a 200 status code).

    Parameters
    ----------
    response : requests.Response
        response object from requests.request().

    Raises
    ------
    InvalidResponseError
        if the response is not valid.
    AuthenticationError
        if the response is a 401 status code.
    KeyboardInterrupt
        if the user interrupts the download.

    Returns
    -------
    dict
        response as a dictionary. If the response is not valid, this function will raise
        an exception.
    """

    if not response.ok:
        logger.info("Response status is NOT OK : %s", response.status_code)
        error_str = format_invalid_response_msg(response, user_id)
        if response.status_code == 401:
            raise AuthenticationError(error_str)

        if HEARTBEAT_ENDPOINT in response.request.url:
            raise HeartbeatError(error_str)

        raise InvalidResponseError(
            f"Request did not return a 200 status code.\n{error_str}"
        )

    try:
        response_dict = response.json()
        if response_dict is None:
            error_str = format_invalid_response_msg(response, user_id)
            raise InvalidResponseError(f"Response is empty.\n{error_str}")
        return response_dict
    except Exception as exc:
        if isinstance(exc, KeyboardInterrupt):
            raise exc

        error_str = format_invalid_response_msg(response, user_id)
        raise InvalidResponseError(error_str + f"Error parsing response as JSON: {exc}")


def format_invalid_response_msg(response: requests.Response, user_id: str) -> str:
    """
    This function formats an error message for an invalid response from the API.
    Should only be called if there is an error in the response (as the functions adds the
    response text to the error message).
    """
    error_str: str = (
        f"Response: {response}\n"
        f"User ID: {user_id}\n"
        f"Requested URL: {response.request.url}\n"
        f"Response status code: {response.status_code}\n"
        f"Response headers: {response.headers}\n"
        f"Response text: {response.text}\n"
        f"DataQuery Interaction ID: {response.headers.get('x-dataquery-interaction-id', 'N/A')}\n"
        f"Timestamp (UTC): {datetime.now(timezone.utc).isoformat()}; \n"
    )

    return error_str


def request_wrapper(
    url: str,
    headers: Optional[Dict] = None,
    params: Optional[Dict] = None,
    method: str = "get",
    tracking_id: Optional[str] = None,
    proxy: Optional[Dict] = None,
    cert: Optional[Tuple[str, str]] = None,
    **kwargs,
) -> dict:
    """
    Wrapper for requests.request() that handles retries and logging. All parameters and
    kwargs are passed to requests.request().

    Parameters
    ----------
    url : str
        URL to request.
    headers : dict
        headers to pass to requests.request().
    params : dict
        params to pass to requests.request().
    method : str
        HTTP method to use. Must be one of "get" or "post". Defaults to "get".
    kwargs : dict
        kwargs to pass to requests.request().
    tracking_id : str
        default None, unique tracking ID of request.
    proxy : dict
        default None, dictionary of proxy settings for request.
    cert : Tuple[str, str]
        default None, tuple of string for filename of certificate and key.

    Raises
    ------
    InvalidResponseError
        if the response is not valid.
    AuthenticationError
        if the response is a 401 status code.
    DownloadError
        if the request fails after retrying.
    KeyboardInterrupt
        if the user interrupts the download.
    ValueError
        if the method is not one of "get" or "post".
    Exception
        other exceptions may be raised by requests.request().

    Returns
    -------
    dict
        response as a dictionary.
    """

    if method not in ["get", "post"]:
        raise ValueError(f"Invalid method: {method}")

    user_id: str = kwargs.pop("user_id", "unknown")

    verify: bool = kwargs.pop("verify", True)

    # insert tracking info in headers
    if headers is None:
        headers: Dict = {}
    headers["User-Agent"] = f"MacrosynergyPackage/{ms_version_info}"

    uuid_str: str = str(uuid.uuid4())
    if (tracking_id is None) or (tracking_id == ""):
        tracking_id: str = uuid_str
    else:
        tracking_id: str = f"uuid::{uuid_str}::{tracking_id}"

    headers["X-Tracking-Id"] = tracking_id

    log_url: str = form_full_url(url, params)
    logger.debug(f"Requesting URL: {log_url} with tracking_id: {tracking_id}")
    raised_exceptions: List[Exception] = []
    error_statements: List[str] = []
    error_statement: str = ""
    retry_count: int = 0
    response: Optional[requests.Response] = None
    while retry_count < API_RETRY_COUNT:
        try:
            prepared_request: requests.PreparedRequest = requests.Request(
                method, url, headers=headers, params=params, **kwargs
            ).prepare()

            with requests.Session().send(
                prepared_request,
                proxies=proxy,
                cert=cert,
                timeout=300,
                verify=verify,
            ) as response:
                if isinstance(response, requests.Response):
                    return validate_response(response=response, user_id=user_id)
                else:
                    raise InvalidResponseError(
                        f"Request did not return a response.\n"
                        f"User ID: {user_id}\n"
                        f"Requested URL: {log_url}\n"
                        f"Timestamp (UTC): {datetime.now(timezone.utc).isoformat()}; \n"
                    )

        except Exception as exc:
            # if keyboard interrupt, raise as usual
            if isinstance(exc, KeyboardInterrupt):
                print("KeyboardInterrupt -- halting download")
                raise exc

            # authentication error, clearly not a transient error
            if isinstance(exc, AuthenticationError):
                raise exc

            error_statement = (
                f"Request to {log_url} failed with error {exc}. "
                f"User ID: {user_id}. "
                f"Retry count: {retry_count}. "
                f"Tracking ID: {tracking_id}"
            )
            raised_exceptions.append(exc)
            error_statements.append(error_statement)

            known_exceptions = KNOWN_EXCEPTIONS + [HeartbeatError]
            # NOTE : HeartBeat is a special case

            # NOTE: exceptions that need the code to break should be caught before this
            # all other exceptions are caught here and retried after a delay

            if any([isinstance(exc, e) for e in known_exceptions]):
                logger.warning(error_statement)
                retry_count += 1
                time.sleep(API_DELAY_PARAM)
            else:
                raise exc

    if isinstance(raised_exceptions[-1], HeartbeatError):
        raise HeartbeatError(error_statement)

    errs_str = "\n\n".join(
        ("\t" + str(e) + " - \n\t\t" + est)
        for e, est in zip(raised_exceptions, error_statements)
    )

    e_str = f"Request to {log_url} failed with error {raised_exceptions[-1]}. \n"
    e_str += "-" * 20 + "\n"
    if isinstance(response, requests.Response):
        e_str += f" Status code: {response.status_code}."
    e_str += (
        f" No longer retrying. Tracking ID: {tracking_id}"
        f"Exceptions raised:\n{errs_str}"
    )

    raise DownloadError(e_str)


class OAuth(object):
    """
    Class for handling OAuth authentication for the DataQuery API.

    Parameters
    ----------
    client_id : str
        client ID for the OAuth application.
    client_secret : str
        client secret for the OAuth application.
    proxy : dict
        proxy to use for requests. Defaults to None.
    token_url : str
        URL for getting OAuth tokens.
    dq_resource_id : str
        resource ID for the JPMaQS Application.

    Raises
    ------
    ValueError
        if any of the parameters are semantically incorrect.
    TypeError
        if any of the parameters are of the wrong type.
    Exception
        other exceptions may be raised by underlying functions.
    """

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        proxy: Optional[dict] = None,
        token_url: str = OAUTH_TOKEN_URL,
        dq_resource_id: str = OAUTH_DQ_RESOURCE_ID,
        **kwargs,
    ):
        logger.debug("Instantiate OAuth pathway to DataQuery")
        vars_types_zip: zip = zip(
            [client_id, client_secret, token_url, dq_resource_id],
            [
                "client_id",
                "client_secret",
                "token_url",
                "dq_resource_id",
            ],
        )

        for varx, namex in vars_types_zip:
            if not isinstance(varx, str):
                raise TypeError(f"{namex} must be a <str> and not {type(varx)}.")

        if not isinstance(proxy, dict) and proxy is not None:
            raise TypeError(f"proxy must be a <dict> and not {type(proxy)}.")

        self.token_url: str = token_url
        self.proxy: Optional[dict] = proxy

        self._stored_token: Optional[dict] = None
        self.token_data = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
            "aud": dq_resource_id,
        }

        self.kwargs = kwargs

    def _valid_token(self) -> bool:
        """
        Method to check if the stored token is valid.

        Returns
        -------
        bool
            True if the token is valid, False otherwise.
        """

        if self._stored_token is None:
            logger.debug("No token stored")
            return False

        created: datetime = self._stored_token["created_at"]  # utc time of creation
        expires: datetime = created + timedelta(
            seconds=self._stored_token["expires_in"] * TOKEN_EXPIRY_BUFFER
        )

        utcnow = datetime.now(timezone.utc)
        is_active: bool = expires > utcnow

        logger.debug(
            "Active token: %s, created: %s, expires: %s, now: %s",
            is_active,
            created,
            expires,
            utcnow,
        )

        return is_active

    def _get_token(self) -> str:
        """
        Method to get a new OAuth token.

        Returns
        -------
        str
            OAuth token.
        """

        if not self._valid_token():
            logger.debug("Request new OAuth token")
            js = request_wrapper(
                url=self.token_url,
                data=self.token_data,
                method="post",
                proxy=self.proxy,
                tracking_id=OAUTH_TRACKING_ID,
                user_id=self._get_user_id(),
                **self.kwargs
            )
            # on failure, exception will be raised by request_wrapper

            # NOTE : use UTC time for token expiry
            self._stored_token: dict = {
                "created_at": datetime.now(timezone.utc),
                "access_token": js["access_token"],
                "expires_in": js["expires_in"],
            }

        return self._stored_token["access_token"]

    def _get_user_id(self) -> str:
        return "OAuth_ClientID - " + self.token_data["client_id"]

    def get_auth(self) -> Dict[str, Union[str, Optional[Tuple[str, str]]]]:
        """
        Returns a dictionary with the authentication information, in the same format as
        the `macrosynergy.download.dataquery.CertAuth.get_auth()` method.
        """

        headers: Dict = {"Authorization": "Bearer " + self._get_token()}
        return {
            "headers": headers,
            "cert": None,
            "user_id": self._get_user_id(),
        }


class CertAuth(object):
    """
    Class for handling certificate based authentication for the DataQuery API.

    Parameters
    ----------
    username : str
        username for the DataQuery API.
    password : str
        password for the DataQuery API.
    crt : str
        path to the certificate file.
    key : str
        path to the key file.

    Raises
    ------
    AssertionError
        if any of the parameters are of the wrong type.
    FileNotFoundError
        if certificate or key file is missing from filesystem.
    Exception
        other exceptions may be raised by underlying functions.
    """

    def __init__(
        self,
        username: str,
        password: str,
        crt: str,
        key: str,
        proxy: Optional[dict] = None,
    ):
        for varx, namex in zip([username, password], ["username", "password"]):
            if not isinstance(varx, str):
                raise TypeError(f"{namex} must be a <str> and not {type(varx)}.")

        auth_str = f"{username:s}:{password:s}"
        self.auth: str = base64.b64encode(auth_str.encode("utf-8")).decode("utf-8")

        # Key and Certificate check
        for varx, namex in zip([crt, key], ["crt", "key"]):
            if not isinstance(varx, str):
                raise TypeError(f"{namex} must be a <str> and not {type(varx)}.")
            if not os.path.isfile(varx):
                raise FileNotFoundError(f"The file '{varx}' does not exist.")
        self.key: str = key
        self.crt: str = crt
        self.username: str = username
        self.password: str = password
        self.proxy: Optional[dict] = proxy

    def get_auth(self) -> Dict[str, Union[str, Optional[Tuple[str, str]]]]:
        """
        Returns a dictionary with the authentication information, in the same format as
        the `macrosynergy.download.dataquery.OAuth.get_auth()` method.
        """

        headers = {"Authorization": f"Basic {self.auth:s}"}
        user_id = "CertAuth_Username - " + self.username
        return {
            "headers": headers,
            "cert": (self.crt, self.key),
            "user_id": user_id,
        }


def validate_download_args(
    expressions: List[str],
    start_date: str,
    end_date: str,
    show_progress: bool,
    endpoint: str,
    calender: str,
    frequency: str,
    conversion: str,
    nan_treatment: str,
    reference_data: str,
    retry_counter: int,
    delay_param: float,
    batch_size: int,
):
    """
    Validate the arguments passed to the `download_data()` method.

    Raises
    ------
    TypeError
        if any of the arguments are of the wrong type.
    ValueError
        if any of the arguments are semantically incorrect.

    Returns
    -------
    bool
        True if all arguments are valid.
    """

    if expressions is None:
        raise ValueError("`expressions` must be a list of strings.")

    if not isinstance(expressions, list):
        raise TypeError("`expressions` must be a list of strings.")

    if not all(isinstance(expr, str) for expr in expressions):
        raise TypeError("`expressions` must be a list of strings.")

    for varx, namex in zip([start_date, end_date], ["start_date", "end_date"]):
        if (varx is None) or not isinstance(varx, str):
            raise TypeError(f"`{namex}` must be a string.")
        if not is_valid_iso_date(varx):
            raise ValueError(
                f"`{namex}` must be a string in the ISO-8601 format (YYYY-MM-DD)."
            )

    if not isinstance(show_progress, bool):
        raise TypeError("`show_progress` must be a boolean.")

    if not isinstance(retry_counter, int):
        raise TypeError("`retry_counter` must be an integer.")

    if not isinstance(delay_param, float):
        raise TypeError("`delay_param` must be a float >=0.2 (seconds).")

    if delay_param < 0.0:
        raise ValueError("`delay_param` must be a float >=0.2 (seconds).")

    if delay_param < 0.2:
        warnings.warn(
            RuntimeWarning(
                f"`delay_param` is too low; DataQuery API may reject requests. "
                f"Minimum recommended value is 0.2 seconds. "
            )
        )

    if not isinstance(batch_size, int):
        raise TypeError("`batch_size` must be an integer.")
    if batch_size < 1:
        raise ValueError("`batch_size` must be an integer >=1.")
    elif batch_size > 20:
        warnings.warn(
            RuntimeWarning(
                f"`batch_size` is too high; DataQuery API's time-series endpoint "
                f"accepts a maximum of 20 expressions per request. "
            )
        )

    vars_types_zip: zip = zip(
        [
            endpoint,
            calender,
            frequency,
            conversion,
            nan_treatment,
            reference_data,
        ],
        [
            "endpoint",
            "calender",
            "frequency",
            "conversion",
            "nan_treatment",
            "reference_data",
        ],
    )
    for varx, namex in vars_types_zip:
        if not isinstance(varx, str):
            raise TypeError(f"`{namex}` must be a string.")

    return True


class DataQueryInterface(object):
    """
    High level interface for the DataQuery API.  When using OAuth authentication:

    Parameters
    ----------
    client_id : str
        client ID for the OAuth application.
    client_secret : str
        client secret for the OAuth application.  When using certificate authentication:
    crt : str
        path to the certificate file.
    key : str
        path to the key file.
    username : str
        username for the DataQuery API.
    password : str
        password for the DataQuery API.
    oauth : bool
        whether to use OAuth authentication. Defaults to True.
    debug : bool
        whether to print debug messages. Defaults to False.
    concurrent : bool
        whether to use concurrent requests. Defaults to True.
    batch_size : int
        default 20, number of expressions to send in a single request. Must be a number
        between 1 and 20 (both included).
    check_connection : bool
        whether to send a check_connection request. Defaults to True.
    base_url : str
        base URL for the DataQuery API. Defaults to OAUTH_BASE_URL if `oauth` is True,
        CERT_BASE_URL otherwise.
    token_url : str
        token URL for the DataQuery API. Defaults to OAUTH_TOKEN_URL.
    suppress_warnings : bool
        whether to suppress warnings. Defaults to True.
    custom_auth : Any
        custom authentication object. When specified oauth must be False and the object
        must have a get_auth method. Defaults to None.

    Raises
    ------
    TypeError
        if any of the parameters are of the wrong type.
    ValueError
        if any of the parameters are semantically incorrect.
    InvalidResponseError
        if the response from the server is not valid.
    DownloadError
        if the download fails to complete after a number of retries.
    HeartbeatError
        if the heartbeat (check connection) fails.
    Exception
        other exceptions may be raised by underlying functions.
    """

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        crt: Optional[str] = None,
        key: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        proxy: Optional[dict] = None,
        oauth: bool = True,
        debug: bool = False,
        batch_size: int = 20,
        check_connection: bool = True,
        base_url: str = OAUTH_BASE_URL,
        token_url: str = OAUTH_TOKEN_URL,
        suppress_warning: bool = True,
        custom_auth=None,
        verify: bool = True,
    ):
        self._check_connection: bool = check_connection
        self.msg_errors: List[str] = []
        self.msg_warnings: List[str] = []
        self.unavailable_expressions: List[str] = []
        self.debug: bool = debug
        self.suppress_warning: bool = suppress_warning
        self.batch_size: int = batch_size

        for varx, namex, typex in [
            (client_id, "client_id", str),
            (client_secret, "client_secret", str),
            (crt, "crt", str),
            (key, "key", str),
            (username, "username", str),
            (password, "password", str),
            (proxy, "proxy", dict),
        ]:
            if not isinstance(varx, typex) and varx is not None:
                raise TypeError(f"{namex} must be a {typex} and not {type(varx)}.")

        self.auth: Optional[Union[CertAuth, OAuth]] = None
        if oauth and not all([client_id, client_secret]):
            warnings.warn(
                "OAuth authentication requested but client ID and/or client secret "
                "not found. Falling back to certificate authentication.",
                UserWarning,
            )
            if not all([username, password, crt, key]):
                raise ValueError(
                    "Certificate credentials not found. "
                    "Check the parameters passed to the DataQueryInterface class."
                )
            else:
                oauth: bool = False

        self.verify: bool = verify

        if oauth:
            self.auth: OAuth = OAuth(
                client_id=client_id,
                client_secret=client_secret,
                token_url=token_url,
                proxy=proxy,
                verify=self.verify
            )
        elif custom_auth is not None:
            self.auth = custom_auth
        else:
            if base_url == OAUTH_BASE_URL:
                base_url: str = CERT_BASE_URL

            self.auth: CertAuth = CertAuth(
                username=username,
                password=password,
                crt=crt,
                key=key,
                proxy=proxy,
            )

        assert self.auth is not None, (
            "Unable to instantiate authentication object. "
            "Check the parameters passed to the DataQueryInterface class."
        )

        self.proxy: Optional[dict] = proxy
        self.base_url: str = base_url

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            logger.error("Exception %s - %s", exc_type, exc_value)
            print(f"Exception: {exc_type} {exc_value}")

    def _get_unavailable_expressions(
        self,
        expected_exprs: List[str],
        dicts_list: List[Dict],
    ) -> List[str]:
        """
        Method to get the expressions that are not available in the response. Looks at
        the dict["attributes"][0]["expression"] field of each dict in the list.

        Parameters
        ----------
        expected_exprs : List[str]
            list of expressions that were requested.
        dicts_list : List[Dict]
            list of dicts to search for the expressions.

        Returns
        -------
        List[str]
            list of expressions that were not found in the dicts.
        """

        found_exprs: List[str] = [
            curr_dict["attributes"][0]["expression"]
            for curr_dict in dicts_list
            if curr_dict["attributes"][0]["time-series"] is not None
        ]
        return list(set(expected_exprs) - set(found_exprs))

    def check_connection(self, verbose=False, raise_error: bool = False) -> bool:
        """
        Check the connection to the DataQuery API using the Heartbeat endpoint.

        Parameters
        ----------
        verbose : bool
            whether to print a message if the heartbeat is successful. Useful for
            debugging. Defaults to False.

        Raises
        ------
        HeartbeatError
            if the heartbeat fails.

        Returns
        -------
        bool
            True if the connection is successful, False otherwise.
        """

        logger.debug(f"Sleep before checking connection - {API_DELAY_PARAM} seconds")
        time.sleep(API_DELAY_PARAM)
        logger.debug("Check if connection can be established to JPMorgan DataQuery")
        js: dict = request_wrapper(
            url=self.base_url + HEARTBEAT_ENDPOINT,
            params={"data": "NO_REFERENCE_DATA"},
            proxy=self.proxy,
            tracking_id=HEARTBEAT_TRACKING_ID,
            verify=self.verify,
            **self.auth.get_auth()
        )

        result: bool = True
        if (js is None) or (not isinstance(js, dict)) or ("info" not in js):
            logger.warning("Connection to JPMorgan DataQuery heartbeat failed")
            result: bool = False

        if result:
            result = (int(js["info"]["code"]) == 200) and (
                js["info"]["message"] == "Service Available."
            )

        if verbose:
            print("Connection successful!" if result else "Connection failed.")
        if raise_error and not result:
            raise ConnectionError(HeartbeatError("Heartbeat failed."))
        return result

    def _fetch(
        self,
        url: str,
        params: dict = None,
        tracking_id: Optional[str] = None,
    ) -> List[Dict]:
        """
        Make a request to the DataQuery API using the specified parameters. Used to wrap
        a request in a thread for concurrent requests, or to simplify the code for
        single requests. Used by the `_fetch_timeseries()` method.

        Parameters
        ----------
        url : str
            URL to request.
        params : dict
            parameters to send with the request.
        proxy : dict
            proxy to use for the request.
        tracking_id : str
            tracking ID to use for the request.

        Raises
        ------
        InvalidResponseError
            if the response from the server is not valid.
        Exception
            other exceptions may be raised by underlying functions.

        Returns
        -------
        List[Dict]
            list of dictionaries containing the response data.
        """

        downloaded_data: List[Dict] = []
        response: Dict = request_wrapper(
            url=url,
            params=params,
            proxy=self.proxy,
            tracking_id=tracking_id,
            verify=self.verify,
            **self.auth.get_auth(),
        )

        if (response is None) or ("instruments" not in response.keys()):
            if response is not None:
                if (
                    ("info" in response)
                    and ("code" in response["info"])
                    and (int(response["info"]["code"]) == 204)
                ):
                    raise NoContentError(
                        f"Content was not found for the request: {response}\n"
                        f"User ID: {self.auth.get_auth()['user_id']}\n"
                        f"URL: {form_full_url(url, params)}\n"
                        f"Timestamp (UTC): {datetime.now(timezone.utc).isoformat()}"
                    )

            raise InvalidResponseError(
                f"Invalid response from DataQuery: {response}\n"
                f"User ID: {self.auth.get_auth()['user_id']}\n"
                f"URL: {form_full_url(url, params)}"
                f"Timestamp (UTC): {datetime.now(timezone.utc).isoformat()}"
            )

        downloaded_data.extend(response["instruments"])

        if "links" in response.keys() and response["links"][1]["next"] is not None:
            logger.debug("DQ response paginated - get next response page")
            downloaded_data.extend(
                self._fetch(
                    url=self.base_url + response["links"][1]["next"],
                    params={},
                    tracking_id=tracking_id,
                )
            )

        return downloaded_data

    def _fetch_timeseries(
        self, url: str, params: dict, tracking_id: str = None, *args, **kwargs
    ) -> List[Dict]:
        """
        Exists to provide a wrapper for the `_fetch()` method that can be modified when
        inheriting from this class. This method is used by the `_concurrent_loop()`
        method.
        """

        return self._fetch(url=url, params=params, tracking_id=tracking_id)

    def get_catalogue(
        self,
        group_id: str = JPMAQS_GROUP_ID,
        page_size: int = 1000,
        verbose: bool = True,
    ) -> List[str]:
        """
        Method to get the JPMaQS catalogue. Queries the DataQuery API's Groups/Search
        endpoint to get the list of tickers in the JPMaQS group. The group ID can be
        changed to fetch a different group's catalogue.

        Parameters
        ----------
        group_id : str
            the group ID to fetch the catalogue for. Defaults to "JPMAQS".
        page_size : int
            the number of tickers to fetch in a single request. Defaults to 1000
            (maximum allowed by the API).

        Raises
        ------
        ValueError
            if the response from the server is not valid.

        Returns
        -------
        List[str]
            list of all tickers in the requested group.
        """

        if not isinstance(group_id, str):
            raise TypeError("`group_id` must be a string.")

        pgsize_err = "`page_size` must be an integer between 1 and 1000."
        if not isinstance(page_size, int):
            raise TypeError(pgsize_err)
        elif (page_size < 1) or (page_size > 1000):
            raise ValueError(pgsize_err)

        if verbose:
            print(f"Downloading the {group_id} catalogue from DataQuery...")
        try:
            response_list: Dict = self._fetch(
                url=self.base_url + CATALOGUE_ENDPOINT,
                params={"group-id": group_id, "limit": page_size},
                tracking_id=CATALOGUE_TRACKING_ID,
            )
        except Exception as e:
            raise e

        tickers: List[str] = [d["instrument-name"] for d in response_list]
        utkr_count: int = len(tickers)
        tkr_idx: List[int] = sorted([d["item"] for d in response_list])

        if not (
            (min(tkr_idx) == 1)
            and (max(tkr_idx) == utkr_count)
            and (len(set(tkr_idx)) == utkr_count)
        ):
            raise ValueError("The downloaded catalogue is corrupt.")
        if verbose:
            print(f"Downloaded {group_id} catalogue with {utkr_count} tickers.")

        return tickers

    def _concurrent_loop(
        self,
        expr_batches: List[List[str]],
        show_progress: bool,
        url: str,
        params: dict,
        tracking_id: str,
        delay_param: float,
        *args,
        **kwargs,
    ) -> Tuple[List[Union[Dict, Any]], List[List[str]]]:
        """
        Concurrent loop to download data from the DataQuery API. Used by the
        `_download()` method.

        Returns
        -------
        Tuple[List[Union[Dict, Any]], List[List[str]]]
            tuple of two lists. The first list contains the downloaded data, and the
            second list contains the failed batches.
        """

        future_objects: List[concurrent.futures.Future] = []
        download_outputs: List[Union[Dict, Any]] = []
        failed_batches: List[List[str]] = []
        last_five_exc: List[Exception] = []
        continuous_failures: int = 0
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for ib, expr_batch in tqdm(
                enumerate(expr_batches),
                desc="Requesting data",
                disable=not show_progress,
                total=len(expr_batches),
            ):
                curr_params: Dict = params.copy()
                curr_params["expressions"] = expr_batch
                try:
                    future_objects.append(
                        executor.submit(
                            self._fetch_timeseries,
                            url=url,
                            params=curr_params,
                            tracking_id=tracking_id,
                            *args,
                            **kwargs,
                        )
                    )
                except Exception as exc:
                    raise exc

                time.sleep(delay_param)

            for ib, future in tqdm(
                enumerate(future_objects),
                desc="Downloading data",
                disable=not show_progress,
                total=len(future_objects),
            ):
                try:
                    if future.exception() is not None:
                        raise future.exception()
                    download_outputs.append(future.result())
                    continuous_failures = 0
                except Exception as exc:
                    if isinstance(exc, (KeyboardInterrupt, AuthenticationError)):
                        executor.shutdown(wait=False, cancel_futures=True)
                        raise exc

                    failed_batches.append(expr_batches[ib])
                    self.msg_errors.append(f"Batch {ib} failed with exception: {exc}")
                    continuous_failures += 1
                    last_five_exc.append(exc)
                    if continuous_failures > MAX_CONTINUOUS_FAILURES:
                        exc_str: str = "\n".join([str(e) for e in last_five_exc])
                        raise DownloadError(
                            f"Failed {continuous_failures} times to download data."
                            f" Last five exceptions: \n{exc_str}"
                        )

                    if self.debug:
                        raise exc

        return download_outputs, failed_batches

    def _chain_download_outputs(
        self,
        download_outputs: List[Union[Dict, Any]],
    ) -> List[Dict]:
        """
        Chain the download outputs from the concurrent loop into a single list. Used by
        the `download_data()` method. Exists to provide a method that can be modified
        when inheriting from this class.

        Parameters
        ----------
        download_outputs : List[Union[Dict, Any]
            list of list of dictionaries/ other objects.

        Returns
        -------
        List[Dict]
            list of dictionaries/other objects.
        """

        return list(itertools.chain.from_iterable(download_outputs))

    def _download(
        self,
        expressions: List[str],
        params: dict,
        url: str,
        tracking_id: str,
        delay_param: float,
        show_progress: bool = False,
        retry_counter: int = 0,
        *args,
        **kwargs,
    ) -> List[dict]:
        """
        Backend method to download data from the DataQuery API. Used by the
        `download_data()` method.
        """

        if 0 < retry_counter < HL_RETRY_COUNT:
            print("Retrying failed downloads. Retry count:", retry_counter)

        if retry_counter > HL_RETRY_COUNT:
            error_str = (f"Failed {retry_counter} times to download data all requested data.\n"
                f"No longer retrying.")
            if len(self.msg_errors) > 0:
                error_str += "\n".join(self.msg_errors)
            raise DownloadError(error_str)

        expr_batches: List[List[str]] = [
            expressions[i : i + self.batch_size]
            for i in range(0, len(expressions), self.batch_size)
        ]

        download_outputs: List[List[Dict]]
        failed_batches: List[List[str]]

        download_outputs, failed_batches = self._concurrent_loop(
            expr_batches=expr_batches,
            show_progress=show_progress,
            url=url,
            params=params,
            tracking_id=tracking_id,
            delay_param=delay_param,
            *args,
            **kwargs,
        )

        if len(failed_batches) > 0:
            flat_failed_batches: List[str] = list(
                itertools.chain.from_iterable(failed_batches)
            )
            logger.warning(
                "Failed batches %d - retry download for %d expressions",
                len(failed_batches),
                len(flat_failed_batches),
            )
            retried_output: List[dict] = self._download(
                expressions=flat_failed_batches,
                params=params,
                url=url,
                tracking_id=tracking_id,
                delay_param=delay_param + 0.1,
                show_progress=show_progress,
                retry_counter=retry_counter + 1,
                *args,
                **kwargs,
            )
            download_outputs.extend(retried_output)

        if retry_counter == 0:
            return self._chain_download_outputs(download_outputs)

        return download_outputs

    def download_data(
        self,
        expressions: List[str],
        start_date: str = "2000-01-01",
        end_date: str = None,
        show_progress: bool = False,
        endpoint: str = TIMESERIES_ENDPOINT,
        calender: str = "CAL_ALLDAYS",
        frequency: str = "FREQ_DAY",
        conversion: str = "CONV_LASTBUS_ABS",
        nan_treatment: str = "NA_NOTHING",
        reference_data: str = "NO_REFERENCE_DATA",
        retry_counter: int = 0,
        delay_param: float = API_DELAY_PARAM,
        batch_size: Optional[int] = None,
        *args,
        **kwargs,
    ) -> List[Dict]:
        """
        Download data from the DataQuery API.

        Parameters
        ----------
        expressions : List[str]
            list of expressions to download.
        start_date : str
            start date for the data in the ISO-8601 format (YYYY-MM-DD).
        end_date : str
            end date for the data in the ISO-8601 format (YYYY-MM-DD).
        show_progress : bool
            whether to show a progress bar for the download.
        endpoint : str
            endpoint to use for the download.
        calender : str
            calendar setting to use for the download.
        frequency : str
            frequency of data points to use for the download.
        conversion : str
            conversion setting to use for the download.
        nan_treatment : str
            NaN treatment setting to use for the download.
        reference_data : str
            reference data to pass to the API kwargs.
        retry_counter : int
            number of times the download has been retried.
        delay_param : float
            delay between requests to the API.

        Raises
        ------
        ValueError
            if any arguments are invalid or semantically incorrect (see
            validate_download_args()).
        DownloadError
            if the download fails.
        ConnectionError(HeartbeatError)
            if the heartbeat fails.
        Exception
            other exceptions may be raised by underlying functions.

        Returns
        -------
        List[Dict]
            list of dictionaries containing the response data.
        """

        tracking_id: str = TIMESERIES_TRACKING_ID
        if end_date is None:
            end_date = datetime.today().strftime("%Y-%m-%d")
            # NOTE : if "future dates" are passed, they must be passed by parent functions
            # see jpmaqs.py

        if batch_size is None:
            batch_size = self.batch_size

        # NOTE : args validated only on first call, not on retries
        # this is because the args can be modified by the retry mechanism
        # (eg. date format)

        expressions = sorted(expressions)

        validate_download_args(
            expressions=expressions,
            start_date=start_date,
            end_date=end_date,
            show_progress=show_progress,
            endpoint=endpoint,
            calender=calender,
            frequency=frequency,
            conversion=conversion,
            nan_treatment=nan_treatment,
            reference_data=reference_data,
            retry_counter=retry_counter,
            delay_param=delay_param,
            batch_size=batch_size,
        )

        self.batch_size = batch_size

        if datetime.strptime(end_date, "%Y-%m-%d") < datetime.strptime(
            start_date, "%Y-%m-%d"
        ):
            wStr = "Start date ({}) is after end-date ({}): swapping them!"
            logger.warning(wStr.format(start_date, end_date))
            warnings.warn(wStr.format(start_date, end_date), UserWarning)
            start_date, end_date = end_date, start_date

        # remove dashes from dates to match DQ format
        start_date: str = start_date.replace("-", "")
        end_date: str = end_date.replace("-", "")

        # check heartbeat before each "batch" of requests
        if self._check_connection:
            if not self.check_connection(verbose=True):
                raise ConnectionError(
                    HeartbeatError(
                        f"Heartbeat failed. Timestamp (UTC):"
                        f" {datetime.now(timezone.utc).isoformat()}\n"
                        f"User ID: {self.auth.get_auth()['user_id']}\n"
                    )
                )
            time.sleep(delay_param)

        logger.info(
            "Download %d expressions from DataQuery from %s to %s",
            len(expressions),
            datetime.strptime(start_date, "%Y%m%d").date(),
            datetime.strptime(end_date, "%Y%m%d").date(),
        )
        params_dict: Dict = {
            "format": "JSON",
            "start-date": start_date,
            "end-date": end_date,
            "calendar": calender,
            "frequency": frequency,
            "conversion": conversion,
            "nan_treatment": nan_treatment,
            "data": reference_data,
        }

        final_output: List[dict] = self._download(
            expressions=expressions,
            params=params_dict,
            url=self.base_url + endpoint,
            tracking_id=tracking_id,
            delay_param=delay_param,
            show_progress=show_progress,
            *args,
            **kwargs,
        )

        if (
            isinstance(final_output, list)
            and (len(final_output) > 0)
            and isinstance(final_output[0], dict)
        ):
            self.unavailable_expressions = self._get_unavailable_expressions(
                expected_exprs=expressions, dicts_list=final_output
            )
            logger.info(
                "Downloaded expressions: %d, unavailable: %d",
                len(final_output),
                len(self.unavailable_expressions),
            )

        return final_output


if __name__ == "__main__":
    import os

    client_id: str = os.getenv("DQ_CLIENT_ID")
    client_secret: str = os.getenv("DQ_CLIENT_SECRET")

    expressions = ["DB(CFX,GBP,)"]

    with DataQueryInterface(
        client_id=client_id,
        client_secret=client_secret,
    ) as dq:
        assert dq.check_connection(verbose=True)

        data = dq.download_data(
            expressions=expressions,
            start_date="2024-01-25",
            end_date="2024-02-05",
            show_progress=True,
        )

    print(data)

    print(f"Succesfully downloaded data for {len(data)} expressions.")



"""Convergence of rows for max weight"""

import numpy as np


class ConvergeRow(object):
    """
    Class designed to receive a row of weights, where at least one weight in the
    aforementioned row exceeds the permitted upper-bound, and subsequently redistributes
    the excess evenly across all cross-sections.

    Parameters
    ----------
    row : ~numpy.ndarray
        Array of weights.
    max_weight : float
        Maximum weight.
    margin : float
        Margin of error allowed in the convergence to within the upper-bound,
        "max_weight".
    max_loops : int
        Controls the accuracy: in theory, the greater the number of loops allowed, the
        more accurate the convergence. However, will only become significant if a tight
        margin is imposed: the "looser" the margin, the less likely the maximum number of
        loops permitted will be exceeded.
    """

    def __init__(
        self,
        row: np.ndarray,
        max_weight: float,
        margin: float = 0.001,
        max_loops: int = 25,
    ):
        self.row = row
        self.max_weight = max_weight
        self.flag = (1 / np.sum(self.row > 0)) <= self.max_weight
        self.margin = margin
        self.max_loops = max_loops

    @classmethod
    def application(cls, row: np.ndarray, max_weight: float):
        """
        Initiates the class and applies the redistribution of weights.

        Parameters
        ----------
        row : ~numpy.ndarray
            Array of weights.
        max_weight : float
            Maximum weight.

        Returns
        -------
        ~numpy.ndarray
            Array of weights.
        """
        cr = ConvergeRow(row=row, max_weight=max_weight)

        if cr.flag:  # if enough non-nan values in row
            cr.distribute_simple()
        else:  # apply equal weights to all non-nan cases
            cr.row = (cr.row > 0) / np.sum(cr.row > 0)
            cr.row[cr.row == 0.0] = np.nan

        return cr.row

    def distribute_simple(self):
        """
        Initiates an indefinite While Loop until the weights converge below the
        (max_weight + margin). Will evenly redistribute the excess weight across all
        active cross-sections, and will compute the maximum weight, the number of cross-
        sections above the threshold and the excess weight dynamically: through each
        iteration.
        """

        count = 0
        close_enough = False

        while (count <= self.max_loops) and (not close_enough):
            count += 1
            excesses = self.row - (self.max_weight + self.margin)
            excesses[excesses <= 0] = 0
            self.row = (self.row - excesses) + np.nanmean(excesses)

            if np.max(self.row) <= (self.max_weight + self.margin):
                close_enough = True



"""
Base class for feature selection on panel data.
"""

import pandas as pd

import warnings

from sklearn.base import BaseEstimator
from sklearn.feature_selection import SelectorMixin
from sklearn.exceptions import NotFittedError

from abc import ABC, abstractmethod


class BasePanelSelector(BaseEstimator, SelectorMixin, ABC):
    """
    Base class for statistical feature selection over a panel.
    """

    def __init__(self):
        self.n_features_in_ = None
        self.feature_names_in_ = None

    def fit(self, X, y=None):
        """
        Learn optimal features based on a training set pair (X, y).

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        y : pandas.Series or pandas.DataFrame, optional
            The target vector.
        """
        # Checks
        self._check_fit_params(X, y)

        # Store useful input information
        self.n_features_in_ = X.shape[1]
        self.feature_names_in_ = X.columns

        # Standardise the features for fair comparison
        X = ((X - X.mean()) / X.std()).copy()

        self.mask = self.determine_features(X, y)

        return self

    @abstractmethod
    def determine_features(self, X, y):
        """
        Determine mask of selected features based on a training set pair (X, y).

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        y : pandas.Series or pandas.DataFrame
            The target vector.

        Returns
        -------
        mask : np.ndarray
            Boolean mask of selected features.
        """
        pass

    def transform(self, X):
        """
        Transform method to return only the selected features of the dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.

        Returns
        -------
        X_transformed : pandas.DataFrame
            The feature matrix with only the selected features.
        """
        # Checks
        self._check_transform_params(X)

        # Return selected features
        if sum(self.mask) == 0:
            # Then no features were selected
            warnings.warn(
                "No features were selected. At the given time, no trading decisions can be made based on these features.",
                RuntimeWarning,
            )
            return X.iloc[:, :0]

        return X.loc[:, self.mask]

    def _get_support_mask(self):
        """
        Private method to return a boolean mask of the features selected for the Pandas
        dataframe.
        """
        return self.mask

    def get_feature_names_out(self):
        """
        Method to mask feature names according to selected features.
        """
        if self.feature_names_in_ is None:
            raise NotFittedError(
                "The selector has not been fitted. Please fit the selector ",
                "before calling get_feature_names_out().",
            )

        return self.feature_names_in_[self.get_support(indices=False)]

    def _check_fit_params(self, X, y):
        """
        Checks the input data for the fit method.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        y : pandas.Series or pandas.DataFrame, optional
            The target vector.
        """
        # Check input feature matrix
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the selector must be a pandas dataframe. ",
                "If used as part of an sklearn pipeline, ensure that previous steps ",
                "return a pandas dataframe.",
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("The input dataframe must be multi-indexed.")
        if not X.index.nlevels == 2:
            raise ValueError("X must be multi-indexed with two levels.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for a panel selector ",
                "must be numeric.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix for a panel selector must not contain any "
                "missing values."
            )
        # Check target vector, if provided
        if y is not None:
            if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
                raise TypeError(
                    "Target vector for the selector must be a pandas series or dataframe. ",
                    "If used as part of an sklearn pipeline, ensure that previous steps ",
                    "return a pandas series or dataframe.",
                )
            if isinstance(y, pd.DataFrame):
                if y.shape[1] != 1:
                    raise ValueError(
                        "The target dataframe must have only one column. If used as part of ",
                        "an sklearn pipeline, ensure that previous steps return a pandas ",
                        "series or single-column dataframe.",
                    )
            if not isinstance(y.index, pd.MultiIndex):
                raise ValueError("The target vector must be multi-indexed.")
            if not y.index.nlevels == 2:
                raise ValueError("y must be multi-indexed with two levels.")
            if not y.index.get_level_values(0).dtype == "object":
                raise TypeError("The outer index of y must be strings.")
            if not y.index.get_level_values(1).dtype == "datetime64[ns]":
                raise TypeError("The inner index of y must be datetime.date.")
            if not X.index.equals(y.index):
                raise ValueError(
                    "The indices of the input dataframe X and the output dataframe y don't "
                    "match."
                )
            if y.isnull().values.any():
                raise ValueError(
                    "The target vector for a panel selector must not contain any "
                    "missing values."
                )

    def _check_transform_params(self, X):
        """
        Input checks for the transform method.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        """
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the selector must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("The input dataframe must be multi-indexed.")
        if not X.index.nlevels == 2:
            raise ValueError("X must be multi-indexed with two levels.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for a panel selector ",
                "must be numeric.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix for a panel selector must not contain any "
                "missing values."
            )

        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "The input feature matrix must have the same number of columns as the "
                "training feature matrix."
            )
        if not X.columns.equals(self.feature_names_in_):
            raise ValueError(
                "The input feature matrix must have the same columns as the training "
                "feature matrix."
            )



"""
Module for plotting ranges of values across cross-sections for one or more categories.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Tuple, Callable, Optional
from packaging import version

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df
from macrosynergy.management.types import QuantamentalDataFrame


def view_ranges(
    df: pd.DataFrame,
    xcats: List[str],
    cids: Optional[List[str]] = None,
    start: str = None,
    end: Optional[str] = None,
    val: str = "value",
    kind: str = "bar",
    sort_cids_by: Optional[str] = None,
    title: Optional[str] = None,
    ylab: Optional[str] = None,
    size: Tuple[float] = (16, 8),
    xcat_labels: Optional[List[str]] = None,
    legend_loc: str = "lower center",
    legend_bbox_to_anchor: Optional[Tuple[float]] = None,
    facet: bool = False,
    ncols: int = None,
    nrows: int = None,
    drop_cid_labels: bool = False,
):
    """
    Plots averages and various ranges across sections for one or more categories.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized DataFrame with the necessary columns: 'cid', 'xcat', 'real_date'
        and at least one column with values of interest.
    xcats : List[str]
        extended categories to be checked on. Default is all in the DataFrame.
    cids : List[str]
        cross sections to plot. Default is all in DataFrame.
    start : str
        earliest date in ISO format. Default earliest date in df.
    end : str
        latest date in ISO format. Default is latest date in df.
    val : str
        name of column that contains the values. Default is 'value'.
    kind : str
        type of range plot. Default is 'bar'; other option is 'box'.
    sort_cids_by : str
        criterion for sorting cids on x-axis; Arguments can be 'mean' and 'std'. Default
        is None, i.e. original order. Ordering will be based on the first category if the
        category is defined over the complete panel. Otherwise, mean and standard deviation
        calculated, of the cross-sections, computed across all categories.
    title : str
        string of chart title; defaults depend on type of range plot.
    ylab : str
        y label. Default is no label.
    size : Tuple[float]
        Tuple of width and height of graph. Default is (16, 8).
    xcat_labels : Union[List[str], Dict[str, str]]
        custom labels to be used for the ranges.
    legend_loc : str
        location of legend; passed to matplotlib.pyplot.legend() as `loc`. Default is
        'center right'.
    legend_bbox_to_anchor : Tuple[float]
        passed to matplotlib.pyplot.legend() as `bbox_to_anchor`. Default is None.
        Please see [Matplotlib's Legend Documentation]:
        (https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html) for more
        information on the legend parameters `loc` and `bbox_to_anchor`.
    facet : bool
        If True, facets the plot by xcat. Default is False.
    ncols : int
        Number of columns in the facet plot. Default is None, which will be set to
        `ceil(n_xcats / nrows)`.
    nrows : int
        Number of rows in the facet plot. Default is None, which will be set to
        `min(n_xcats, 2)`.
    drop_cid_labels : bool
        If True, the x-axis labels for cids will be dropped when plotting a facet. 
        Default is False. This is useful when there are many cids and the labels would 
        overlap.
    """

    df = QuantamentalDataFrame(df)

    possible_xcats = set(df["xcat"])
    missing_xcats = set(xcats).difference(possible_xcats)
    error_xcats = (
        "The categories passed in to view_ranges() must be present in the "
        f"DataFrame: missing {missing_xcats}."
    )
    if not set(xcats).issubset(possible_xcats):
        raise ValueError(error_xcats)

    sort_cids_func: Callable = None
    if sort_cids_by is not None:
        if not isinstance(sort_cids_by, str):
            raise TypeError("`sort_cids_by` must be a string.")
        sort_error = "Sorting parameter must either be 'mean' or 'std'."
        if  sort_cids_by not in ["mean", "std"]:
            raise ValueError(sort_error)
        if sort_cids_by == "mean":
            sort_cids_func = np.mean
        else:
            sort_cids_func = np.std

    error_message = (
        "The number of custom labels must match the defined number of "
        "categories in pnl_cats."
    )
    if xcat_labels is not None:
        if isinstance(xcat_labels, dict):
            xcat_labels = [xcat_labels[xcat] for xcat in xcats]
        if len(xcat_labels) != len(xcats):
            raise ValueError(error_message)

    # Unique cross-sections across the union of categories passed - not the intersection.
    # Order of categories will be preserved.
    df, xcats, cids = reduce_df(df, xcats, cids, start, end, out_all=True)

    s_date = df["real_date"].min().strftime("%Y-%m-%d")
    e_date = df["real_date"].max().strftime("%Y-%m-%d")

    sns.set(style="darkgrid")

    if title is None:
        if kind == "bar":
            title = f"Means and standard deviations from {s_date} to {e_date}."
        elif kind == "box":
            title = (
                f"Interquartile ranges, extended ranges and outliers "
                f"from {s_date} to {e_date}."
            )
    if ylab is None:
        ylab = ""

    filt_1 = df["xcat"] == xcats[0]
    first_xcat_cids = set(df[filt_1]["cid"])
    # First category is not defined over all cross-sections.
    order_condition = list(set(cids)) == list(first_xcat_cids)

    if order_condition and sort_cids_func is not None:
        # Sort exclusively on the first category.
        dfx = df[filt_1].groupby(["cid"], observed=True)[val].apply(sort_cids_func)
        order = dfx.sort_values(ascending=False).index
    elif not order_condition and sort_cids_func is not None:
        # Sort across all categories on the available cross-sections.
        dfx = df.groupby(["cid"], observed=True)[val].apply(sort_cids_func)
        order = dfx.sort_values(ascending=False).index
    else:
        order = None

    sns.set_theme(style="darkgrid")

    def _plot_bar(ax, data, order, val, xcat=None):
        if version.parse(sns.__version__) >= version.parse("0.12.0"):
            return sns.barplot(
                x="cid",
                y=val,
                hue="xcat" if xcat is None else None,
                hue_order=xcats if xcat is None else None,
                palette="Paired",
                data=data,
                errorbar="sd",
                order=order,
                ax=ax if xcat is not None else None,
            )
        else:
            return sns.barplot(
                x="cid",
                y=val,
                hue="xcat" if xcat is None else None,
                hue_order=xcats if xcat is None else None,
                palette="Paired",
                data=data,
                ci="sd",
                order=order,
                ax=ax if xcat is not None else None,
            )

    def _plot_box(ax, data, order, val, xcat=None):
        return sns.boxplot(
            x="cid",
            y=val,
            hue="xcat" if xcat is None else None,
            hue_order=xcats if xcat is None else None,
            palette="Paired",
            data=data,
            order=order,
            ax=ax if xcat is not None else None,
        )

    def _set_facet_axis(ax, title, ylab):
        ax.set_title(title, fontdict={"fontsize": 14})
        ax.set_xlabel("")
        ax.set_ylabel(ylab)
        ax.xaxis.grid(True)
        ax.axhline(0, ls="--", linewidth=1, color="black")
        if ax.get_legend() is not None:
            ax.get_legend().remove()
        if drop_cid_labels:
            ax.set_xticklabels([])
            ax.set_xlabel("")

            ax.set_xticklabels([])

    def _set_main_axis(ax, title, ylab):
        ax.set_title(title, fontdict={"fontsize": 16})
        ax.set_xlabel("")
        ax.set_ylabel(ylab)
        ax.xaxis.grid(True)
        ax.axhline(0, ls="--", linewidth=1, color="black")

    if facet:
        n_xcats = len(xcats)
        if ncols is None and nrows is None:
            max_rows = 2
            nrows = min(n_xcats, max_rows)
            ncols = int(np.ceil(n_xcats / nrows))
        figsize = (size[0] * ncols / max(1, n_xcats/2), size[1] * nrows / max(1, n_xcats/2))
        fig, axes = plt.subplots(nrows, ncols, figsize=figsize, squeeze=False)
        axes = axes.flatten()
        for i, xcat in enumerate(xcats):
            ax = axes[i]
            df_xcat = df[df["xcat"] == xcat]
            if kind == "bar":
                _plot_bar(ax, df_xcat, order, val, xcat=xcat)
            elif kind == "box":
                _plot_box(ax, df_xcat, order, val, xcat=xcat)
            facet_title = xcat_labels[i] if xcat_labels is not None else xcat
            _set_facet_axis(ax, facet_title, ylab)
        for j in range(i + 1, len(axes)):
            fig.delaxes(axes[j])
        fig.suptitle(title, fontsize=16)
        plt.tight_layout(rect=[0, 0, 1, 0.97])
        plt.show()
    else:
        fig, ax = plt.subplots(figsize=size)
        if kind == "bar":
            _plot_bar(ax, df, order, val)
        elif kind == "box":
            _plot_box(ax, df, order, val)
        _set_main_axis(ax, title, ylab)
        handles, labels = ax.get_legend_handles_labels()
        if xcat_labels is not None:
            labels = xcat_labels
        if (len(xcats) == 1) and (xcat_labels is None):
            if ax.get_legend() is not None:
                ax.get_legend().remove()
        else:
            if legend_bbox_to_anchor is None:
                legend_bbox_to_anchor = (0.5, -0.15 - 0.05 * (len(xcats) - 2))
            ax.legend(
                handles=handles[0:],
                labels=labels[0:],
                loc=legend_loc,
                bbox_to_anchor=legend_bbox_to_anchor,
            )
        plt.tight_layout()
        plt.show()


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD",] = ["2010-01-01", "2020-12-31", 0.5, 0.2]
    df_cids.loc["CAD",] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP",] = ["2012-01-01", "2020-11-30", 0, 2]
    df_cids.loc["USD",] = ["2012-01-01", "2020-11-30", 1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR",] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["XR2",] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["XR3",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["XR4",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["XR5",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["CRY",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["INFL",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["GROWTH",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]


    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    view_ranges(
        dfd,
        xcats=["XR"],
        kind="box",
        start="2012-01-01",
        end="2018-01-01",
        sort_cids_by="std",
    )

    filter_1 = (dfd["xcat"] == "XR") & (dfd["cid"] == "AUD")
    dfd = dfd[~filter_1]

    view_ranges(
        dfd,
        xcats=["XR", "CRY", "INFL", "GROWTH"],
        cids=cids,
        kind="box",
        start="2012-01-01",
        end="2018-01-01",
        sort_cids_by=None,
        # xcat_labels=["EQXR_NSA"],
    )
    view_ranges(
        dfd,
        xcats=["XR", "CRY", "INFL", "GROWTH", "XR2", "XR3"],
        cids=cids,
        kind="box",
        start="2012-01-01",
        end="2018-01-01",
        sort_cids_by=None,
        facet=True,
        ncols=2,
        nrows=3,
        drop_cid_labels=True,
    )



import numbers
import numpy as np
import pandas as pd

from sklearn.linear_model import LinearRegression
from macrosynergy.learning.forecasting.bootstrap import (
    BaseModifiedRegressor,
)
from macrosynergy.learning.forecasting.linear_model.ls_regressors.weighted_ls_regressors import (
    SignWeightedLinearRegression,
    TimeWeightedLinearRegression,
)


class ModifiedLinearRegression(BaseModifiedRegressor):
    def __init__(
        self,
        method,
        fit_intercept=True,
        positive=False,
        error_offset=1e-2,
        bootstrap_method="panel",
        bootstrap_iters=1000,
        resample_ratio=1,
        analytic_method=None,
    ):
        """
        Modified OLS linear regression model. Estimated coefficients are divided
        by estimated standard errors to form an auxiliary factor model.

        Parameters
        ----------
        method : str
            Method to modify coefficients. Accepted values are
            "analytic" or "bootstrap".
        fit_intercept : bool, default = True
            Whether to fit an intercept term in the model. Default is True.
        positive : bool, default = False
            Whether to constrain the coefficients to be positive. Default is False.
        error_offset : float, default = 1e-2
            Small offset to add to estimated standard errors in order to prevent
            small denominators during the coefficient adjustment.
        bootstrap_method : str, default = "panel"
            Method used to modify coefficients, when `method = bootstrap`.
            Accepted values are "panel", "period", "cross", "cross_per_period",
            "period_per_cross".
        bootstrap_iters : int, default = 1000
            Number of bootstrap iterations to determine standard errors, used
            only when `method = bootstrap`.
        resample_ratio : numbers.Number, default = 1
            Ratio of resampling units in each bootstrap dataset, used only
            when `method = bootstrap`. This is a fraction of the quantity of
            the panel component to be resampled.
        analytic_method : str, default = None
            The analytic method used to determine standard errors. If `method = analytic`,
            the default standard error expressions for an OLS linear regression model
            are used. If `analytic_method = "White"`, the heteroskedasticity-robust
            White estimator is used to estimate the standard errors.

        Notes
        -----
        OLS linear regression models are fit by finding optimal parameters that
        minimize the total squared residuals. In the frequentist statistics framework,
        "true" population-wide values exist for these parameters, which can only be
        estimated from sampled data. Consequently, our parameter estimates can be
        considered to be realizations from a random variable, and hence subject to
        sampling variation. Broadly speaking, the greater the amount of independent data
        sampled, the smaller the variation in parameter estimates. In other words,
        parameter estimates are more unreliable when less data is seen during training.
        By estimating the standard deviation of their sampling distributions - a.k.a.
        their "standard errors" - we can adjust our model coefficients to account for
        lack of statistical precision.

        In our `ModifiedLinearRegression`, each estimated parameter is divided by the
        estimated standard error (plus an offset). This means that greater
        volatility in a parameter estimate due to lack of data is accounted for by
        reducing the magnitude of this estimate, whilst greater certainty in the precision
        of the estimate is reflected by inflating a regression coefficient.

        This procedure works for linear models because they're interpretable, with the
        coefficient adjustment having the interpretation of increasing the relevance of
        factors whose coefficients we are more confident in, and decreasing relevance for
        factors whose coefficients we are less confident in.
        """
        # Checks
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")
        self.fit_intercept = fit_intercept
        self.positive = positive

        super().__init__(
            model=LinearRegression(
                fit_intercept=self.fit_intercept, positive=self.positive
            ),
            method=method,
            error_offset=error_offset,
            bootstrap_method=bootstrap_method,
            bootstrap_iters=bootstrap_iters,
            resample_ratio=resample_ratio,
            analytic_method=analytic_method,
        )

    def adjust_analytical_se(
        self,
        model,
        X,
        y,
        analytic_method=None,
    ):
        """
        Adjust the coefficients of the OLS linear regression model
        by an analytical standard error formula.

        Parameters
        ----------
        model : LinearRegression
            The underlying OLS linear regression model to be modified.
        X : pd.DataFrame
            Input feature matrix.
        y : pd.DataFrame or pd.Series
            Target vector associated with each sample in X.
        analytic_method : str, default = None
            The analytic method used to calculate standard errors.

        Returns
        -------
        intercept : float
            Adjusted intercept.
        coef : np.ndarray
            Adjusted coefficients.

        Notes
        -----
        By default, the calculated standard errors use the usual standard
        error expression for OLS linear regression models under the assumption
        of multivariate normality, homoskedasticity and zero mean of the model errors.
        If `analytic_method = "White"`, the HC3 White estimator is used.

        References
        ----------
        [1] https://online.stat.psu.edu/stat462/node/131/
        [2] https://en.wikipedia.org/wiki/Heteroskedasticity-consistent_standard_errors
        """
        # Checks
        if analytic_method is not None:
            if not isinstance(analytic_method, str):
                raise TypeError("analytic_method must be a string.")
            if analytic_method not in ["White"]:
                raise ValueError("analytic_method must be 'White'.")

        if self.fit_intercept:
            X_new = np.column_stack((np.ones(len(X)), X.values))
        else:
            X_new = X.values

        # Calculate the standard errors
        predictions = model.predict(X)
        residuals = (y - predictions).to_numpy()
        XtX_inv = np.linalg.inv(X_new.T @ X_new)
        if analytic_method is None:
            se = np.sqrt(
                np.diag(
                    XtX_inv
                    * np.sum(np.square(residuals))
                    / (X_new.shape[0] - X_new.shape[1])
                )
            )

        elif analytic_method == "White":
            # Implement HC3
            leverages = np.sum((X_new @ XtX_inv) * X_new, axis=1)
            weights = 1 / (1 - leverages) ** 2
            residuals_squared = np.square(residuals)
            weighted_residuals_squared = weights * residuals_squared
            Omega = X_new.T * weighted_residuals_squared @ X_new
            cov_matrix = XtX_inv @ Omega @ XtX_inv
            se = np.sqrt(np.diag(cov_matrix))

        else:
            raise NotImplementedError(
                "Currently, only the standard and HC3 White standard errors are implemented"
            )

        if self.fit_intercept:
            coef_se = se[1:]
            intercept_se = se[0]
        else:
            coef_se = se
            intercept_se = 0

        # Adjust the coefficients and intercepts by the standard errors
        coef = model.coef_ / (coef_se + self.error_offset)

        intercept = model.intercept_ / (intercept_se + self.error_offset)

        return intercept, coef

    def set_params(self, **params):
        """
        Setter method to update the parameters of the ModifiedLinearRegression

        Parameters
        ----------
        **params : dict
            Dictionary of parameters to update.

        Returns
        -------
        self
            The ModifiedLinearRegression instance with updated parameters.
        """
        super().set_params(**params)
        if "fit_intercept" in params or "positive" in params:
            # Re-initialize the LinearRegression instance with updated parameters
            self.model = LinearRegression(
                fit_intercept=self.fit_intercept, positive=self.positive
            )

        return self


class ModifiedSignWeightedLinearRegression(BaseModifiedRegressor):
    def __init__(
        self,
        method,
        fit_intercept=True,
        positive=False,
        error_offset=1e-2,
        bootstrap_method="panel",
        bootstrap_iters=1000,
        resample_ratio=1,
        analytic_method=None,
    ):
        """
        Modified SWLS linear regression model. Estimated coefficients are divided
        by estimated standard errors to form an auxiliary factor model.

        Parameters
        ----------
        method : str
            Method to modify coefficients. Accepted values are
            "analytic" or "bootstrap".
        fit_intercept : bool, default = True
            Whether to fit an intercept term in the model. Default is True.
        positive : bool, default = False
            Whether to constrain the coefficients to be positive. Default is False.
        error_offset : float, default = 1e-2
            Small offset to add to estimated standard errors in order to prevent
            small denominators during the coefficient adjustment.
        bootstrap_method : str, default = "panel"
            Method used to modify coefficients, when `method = bootstrap`.
            Accepted values are "panel", "period", "cross", "cross_per_period",
            "period_per_cross".
        bootstrap_iters : int, default = 1000
            Number of bootstrap iterations to determine standard errors, used
            only when `method = bootstrap`.
        resample_ratio : numbers.Number, default = 1
            Ratio of resampling units in each bootstrap dataset, used only
            when `method = bootstrap`. This is a fraction of the quantity of
            the panel component to be resampled.
        analytic_method : str, default = None
            The analytic method used to determine standard errors. If `method = analytic`,
            the default standard error expressions for an OLS linear regression model
            are used. If `analytic_method = "White"`, the heteroskedasticity-robust
            White estimator is used to estimate the standard errors.

        Notes
        -----
        SWLS linear regression models are fit by finding optimal parameters that
        minimize the total sign-weighted squared residuals. In the frequentist statistics
        framework, "true" population-wide values exist for these parameters, which can
        only be estimated from sampled data. Consequently, our parameter estimates can be
        considered to be realizations from a random variable, and hence subject to
        sampling variation. Broadly speaking, the greater the amount of independent data
        sampled, the smaller the variation in parameter estimates. In other words,
        parameter estimates are more unreliable when less data is seen during training.
        By estimating the standard deviation of their sampling distributions - a.k.a.
        their "standard errors" - we can adjust our model coefficients to account for
        lack of statistical precision.

        In our `ModifiedSignWeightedLinearRegression`, each estimated parameter is divided
        by the estimated standard error (plus an offset). This means that greater
        volatility in a parameter estimate due to lack of data is accounted for by
        reducing the magnitude of this estimate, whilst greater certainty in the precision
        of the estimate is reflected by inflating a regression coefficient.

        This procedure works for linear models because they're interpretable, with the
        coefficient adjustment having the interpretation of increasing the relevance of
        factors whose coefficients we are more confident in, and decreasing relevance for
        factors whose coefficients we are less confident in.
        """
        self.fit_intercept = fit_intercept
        self.positive = positive

        super().__init__(
            model=SignWeightedLinearRegression(
                fit_intercept=self.fit_intercept, positive=self.positive
            ),
            method=method,
            error_offset=error_offset,
            bootstrap_method=bootstrap_method,
            bootstrap_iters=bootstrap_iters,
            resample_ratio=resample_ratio,
            analytic_method=analytic_method,
        )

    def adjust_analytical_se(
        self,
        model,
        X,
        y,
        analytic_method=None,
    ):
        r"""
        Adjust the coefficients of the SWLS linear regression model
        by an analytical standard error formula.

        Parameters
        ----------
        model : SignWeightedLinearRegression
            The underlying SWLS linear regression model to be modified.
        X : pd.DataFrame
            Input feature matrix.
        y : pd.DataFrame or pd.Series
            Target vector associated with each sample in X.
        analytic_method : str, default = None
            The analytic method used to calculate standard errors.

        Returns
        -------
        intercept : float
            Adjusted intercept.
        coef : np.ndarray
            Adjusted coefficients.

        Notes
        ----------
        The analytical parameter estimates for WLS are:

        .. math::

            \hat{\beta}^{\text{WLS}} = (X^{\intercal}WX)^{-1}X^{\intercal}y

        where:
            - `X` is the input feature matrix, possibly with a column of ones representing the choice of an intercept.
            - `W` is the positive-definite, symmetric weight matrix, a diagonal matrix with sample weights along the main diagonal.
            - `y` is the dependent variable vector.

        Since `W` is a positive-definite, symmetric matrix, it has a square root
        equal to the diagonal matrix with square roots of the sample weights along
        the diagonal. Hence, the WLS estimator can be rewritten as:

        .. math::

            \hat{\beta}^{\text{WLS}} = ((({W^{1/2}X})^{\intercal}(W^{1/2}X))^{-1}(W^{1/2}X)^{\intercal}(W^{1/2}y))

        This is precisely the OLS estimator for a rescaled matrix

        .. math::

            \tilde {X} = W^{1/2}X

        and a rescaled dependent variable

        .. math::

            \tilde {y} = W^{1/2}y

        Hence, the usual standard error estimate and White's estimator can be applied
        based on a rescaling of the design matrix and associated target vector.
        """
        # Checks
        if analytic_method is not None:
            if not isinstance(analytic_method, str):
                raise TypeError("analytic_method must be a string.")
            if analytic_method not in ["White"]:
                raise ValueError("analytic_method must be 'White'.")

        if self.fit_intercept:
            X_new = np.column_stack((np.ones(len(X)), X.values))
        else:
            X_new = X.values

        # Get model weights
        weights = model.sample_weights
        # Rescale features and targets by the sign-weighted linear regression sample weights
        X_new = np.sqrt(weights[:, np.newaxis]) * X_new
        y_new = np.sqrt(weights) * y

        # Calculate the standard errors
        predictions = model.predict(X)
        residuals = (y - predictions).to_numpy() * np.sqrt(weights)
        XtX_inv = np.linalg.inv(X_new.T @ X_new)
        if analytic_method is None:
            se = np.sqrt(
                np.diag(
                    XtX_inv
                    * np.sum(np.square(residuals))
                    / (X_new.shape[0] - X_new.shape[1])
                )
            )

        elif analytic_method == "White":
            # Implement HC3
            leverages = np.sum((X_new @ XtX_inv) * X_new, axis=1)
            weights = 1 / (1 - leverages) ** 2
            residuals_squared = np.square(residuals)
            weighted_residuals_squared = weights * residuals_squared
            Omega = X_new.T * weighted_residuals_squared @ X_new
            cov_matrix = XtX_inv @ Omega @ XtX_inv
            se = np.sqrt(np.diag(cov_matrix))

        else:
            raise NotImplementedError(
                "Currently, only the standard and White standard errors are implemented"
            )

        if self.fit_intercept:
            coef_se = se[1:]
            intercept_se = se[0]
        else:
            coef_se = se
            intercept_se = 0

        # Adjust the coefficients and intercepts by the standard errors
        coef = model.coef_ / (coef_se + self.error_offset)
        intercept = model.intercept_ / (intercept_se + self.error_offset)

        return intercept, coef

    def set_params(self, **params):
        """
        Setter method to update the parameters of the
        ModifiedSignWeightedLinearRegression.

        Parameters
        ----------
        **params : dict
            Dictionary of parameters to update.

        Returns
        -------
        self
            The ModifiedSignWeightedLinearRegression instance with updated parameters.
        """
        super().set_params(**params)
        if "fit_intercept" in params or "positive" in params:
            # Re-initialize the SignWeightedLinearRegression instance with updated parameters
            self.model = SignWeightedLinearRegression(
                fit_intercept=self.fit_intercept, positive=self.positive
            )

        return self


class ModifiedTimeWeightedLinearRegression(BaseModifiedRegressor):
    def __init__(
        self,
        method,
        fit_intercept=True,
        positive=False,
        half_life=252,
        error_offset=1e-2,
        bootstrap_method="panel",
        bootstrap_iters=1000,
        resample_ratio=1,
        analytic_method=None,
    ):
        """
        Modified TWLS linear regression model. Estimated coefficients are divided
        by estimated standard errors to form an auxiliary factor model.

        Parameters
        ----------
        method : str
            Method to modify coefficients. Accepted values are
            "analytic" or "bootstrap".
        fit_intercept : bool, default = True
            Whether to fit an intercept term in the model. Default is True.
        positive : bool, default = False
            Whether to constrain the coefficients to be positive. Default is False.
        half_life : int, default = 252
            The half-life of the exponential weighting function
            used to calculate the sample weights.
        error_offset : float, default = 1e-2
            Small offset to add to estimated standard errors in order to prevent
            small denominators during the coefficient adjustment.
        bootstrap_method : str, default = "panel"
            Method used to modify coefficients, when `method = bootstrap`.
            Accepted values are "panel", "period", "cross", "cross_per_period",
            "period_per_cross".
        bootstrap_iters : int, default = 1000
            Number of bootstrap iterations to determine standard errors, used
            only when `method = bootstrap`.
        resample_ratio : numbers.Number, default = 1
            Ratio of resampling units in each bootstrap dataset, used only
            when `method = bootstrap`. This is a fraction of the quantity of
            the panel component to be resampled.
        analytic_method : str, default = None
            The analytic method used to determine standard errors. If `method = analytic`,
            the default standard error expressions for an OLS linear regression model
            are used. If `analytic_method = "White"`, the heteroskedasticity-robust
            White estimator is used to estimate the standard errors.

        Notes
        -----
        TWLS linear regression models are fit by finding optimal parameters that
        minimize the total time-weighted squared residuals. In the frequentist statistics
        framework, "true" population-wide values exist for these parameters, which can
        only be estimated from sampled data. Consequently, our parameter estimates can be
        considered to be realizations from a random variable, and hence subject to
        sampling variation. Broadly speaking, the greater the amount of independent data
        sampled, the smaller the variation in parameter estimates. In other words,
        parameter estimates are more unreliable when less data is seen during training.
        By estimating the standard deviation of their sampling distributions - a.k.a.
        their "standard errors" - we can adjust our model coefficients to account for
        lack of statistical precision.

        In our `ModifiedTimeWeightedLinearRegression`, each estimated parameter is divided
        by the estimated standard error (plus an offset). This means that greater
        volatility in a parameter estimate due to lack of data is accounted for by
        reducing the magnitude of this estimate, whilst greater certainty in the precision
        of the estimate is reflected by inflating a regression coefficient.

        This procedure works for linear models because they're interpretable, with the
        coefficient adjustment having the interpretation of increasing the relevance of
        factors whose coefficients we are more confident in, and decreasing relevance for
        factors whose coefficients we are less confident in.
        """
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.half_life = half_life

        super().__init__(
            model=TimeWeightedLinearRegression(
                fit_intercept=self.fit_intercept,
                positive=self.positive,
                half_life=self.half_life,
            ),
            method=method,
            error_offset=error_offset,
            bootstrap_method=bootstrap_method,
            bootstrap_iters=bootstrap_iters,
            resample_ratio=resample_ratio,
            analytic_method=analytic_method,
        )

    def adjust_analytical_se(
        self,
        model,
        X,
        y,
        analytic_method,
    ):
        r"""
        Adjust the coefficients of the TWLS linear regression model
        by an analytical standard error formula.

        Parameters
        ----------
        model : TimeWeightedLinearRegression
            The underlying TWLS linear regression model to be modified.
        X : pd.DataFrame
            Input feature matrix.
        y : pd.DataFrame or pd.Series
            Target vector associated with each sample in X.
        analytic_method : str, default = None
            The analytic method used to calculate standard errors.

        Returns
        -------
        intercept : float
            Adjusted intercept.
        coef : np.ndarray
            Adjusted coefficients.

        Notes
        ----------
        The analytical parameter estimates for WLS are:

        .. math::

            \hat{\beta}^{\text{WLS}} = (X^{\intercal}WX)^{-1}X^{\intercal}y

        where:
            - `X` is the input feature matrix, possibly with a column of ones representing the choice of an intercept.
            - `W` is the positive-definite, symmetric weight matrix, a diagonal matrix with sample weights along the main diagonal.
            - `y` is the dependent variable vector.

        Since `W` is a positive-definite, symmetric matrix, it has a square root
        equal to the diagonal matrix with square roots of the sample weights along
        the diagonal. Hence, the WLS estimator can be rewritten as:

        .. math::

            \hat{\beta}^{\text{WLS}} = (({W^{1/2}X})^{\intercal}(W^{1/2}X))^{-1}(W^{1/2}X)^{\intercal}(W^{1/2}y))

        This is precisely the OLS estimator for a rescaled matrix

        .. math::
        
            \tilde {X} = W^{1/2}X

        and a rescaled dependent variable

        .. math::
        
            \tilde {y} = W^{1/2}y

        Hence, the usual standard error estimate and White's estimator can be applied
        based on a rescaling of the design matrix and associated target vector.
        """
        # Checks
        if analytic_method is not None:
            if not isinstance(analytic_method, str):
                raise TypeError("analytic_method must be a string.")
            if analytic_method not in ["White"]:
                raise ValueError("analytic_method must be 'White'.")

        if self.fit_intercept:
            X_new = np.column_stack((np.ones(len(X)), X.values))
        else:
            X_new = X.values

        # Get model weights
        weights = model.sample_weights
        # Rescale features and targets by the sign-weighted linear regression sample weights
        X_new = np.sqrt(weights[:, np.newaxis]) * X_new
        y_new = np.sqrt(weights) * y

        # Calculate the standard errors
        predictions = model.predict(X)
        residuals = (y - predictions).to_numpy() * np.sqrt(weights)
        XtX_inv = np.linalg.inv(X_new.T @ X_new)
        if analytic_method is None:
            se = np.sqrt(
                np.diag(
                    XtX_inv
                    * np.sum(np.square(residuals))
                    / (X_new.shape[0] - X_new.shape[1])
                )
            )

        elif analytic_method == "White":
            # Implement HC3
            leverages = np.sum((X_new @ XtX_inv) * X_new, axis=1)
            weights = 1 / (1 - leverages) ** 2
            residuals_squared = np.square(residuals)
            weighted_residuals_squared = weights * residuals_squared
            Omega = X_new.T * weighted_residuals_squared @ X_new
            cov_matrix = XtX_inv @ Omega @ XtX_inv
            se = np.sqrt(np.diag(cov_matrix))

        else:
            raise NotImplementedError(
                "Currently, only the standard and White standard errors are implemented"
            )

        if self.fit_intercept:
            coef_se = se[1:]
            intercept_se = se[0]
        else:
            coef_se = se
            intercept_se = 0

        # Adjust the coefficients and intercepts by the standard errors
        coef = model.coef_ / (coef_se + self.error_offset)
        intercept = model.intercept_ / (intercept_se + self.error_offset)

        return intercept, coef

    def set_params(self, **params):
        """
        Setter method to update the parameters of the
        ModifiedTimeWeightedLinearRegression.

        Parameters
        ----------
        **params : dict
            Dictionary of parameters to update.

        Returns
        -------
        self
            The ModifiedTimeWeightedLinearRegression instance with updated parameters.
        """
        super().set_params(**params)
        if "fit_intercept" in params or "positive" in params or "half_life" in params:
            # Re-initialize the SignWeightedLinearRegression instance with updated parameters
            self.model = TimeWeightedLinearRegression(
                fit_intercept=self.fit_intercept,
                positive=self.positive,
                half_life=self.half_life,
            )

        return self


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_qdf
    import macrosynergy.management as msm
    from macrosynergy.learning import SignalOptimizer, ExpandingKFoldPanelSplit

    from sklearn.linear_model import LinearRegression
    from sklearn.model_selection import GridSearchCV
    from sklearn.metrics import make_scorer, r2_score

    # Randomly generate an unbalanced panel dataset, multi-indexed by cross-section and
    # real_date

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0, 1, 0, 3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 0, 1, 0, 0]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 0, 1, -0.9, 0]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", 0, 1, 0.8, 0]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd = msm.reduce_df(df=dfd, cids=cids, xcats=xcats, blacklist=black)

    dfd = dfd.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    X = dfd.drop(columns=["XR"])
    y = dfd["XR"]

    # First instantiate the BaseModifiedRegressor
    obj = BaseModifiedRegressor(model=LinearRegression(), method="bootstrap")

    # Demonstrate ModifiedLinearRegression usage
    method_pairs = [
        ("analytic", "panel", None),
        ("analytic", "panel", "White"),
        ("bootstrap", "panel", None),
        ("bootstrap", "period", None),
        ("bootstrap", "cross", None),
        ("bootstrap", "cross_per_period", None),
        ("bootstrap", "period_per_cross", None),
    ]
    for method in method_pairs:
        model = ModifiedLinearRegression(
            method=method[0],
            bootstrap_method=method[1],
            bootstrap_iters=100,
            resample_ratio=0.75,
            analytic_method=method[2],
        )
        # Fit the model
        model.fit(X, y)
        print("----")
        print("Modified OLS method:", method)
        print("Modified OLS intercept:", model.intercept_)
        print("Modified OLS coefficients:", model.coef_)
        print("Modified OLS predictions:", model.predict(X))
        print("Modified OLS signal:", model.create_signal(X))
        print("----")

        # Grid search
        cv = GridSearchCV(
            estimator=model,
            param_grid={
                "fit_intercept": [True, False],
                "positive": [True, False],
            },
            cv=3,
            n_jobs=-1,
        )
        cv.fit(X, y)
        print("----")
        print("Modified OLS grid search results:")
        print(pd.DataFrame(cv.cv_results_))
        print("----")



from .jpmaqs import JPMaQSDownload
from .dataquery import DataQueryInterface
from .jpmaqs import custom_download
from .external_data_transformer import transform_to_qdf

__all__ = ["JPMaQSDownload", "DataQueryInterface", "custom_download", "transform_to_qdf"]
